from __future__ import annotations

import random
import uuid

import pandas as pd
import pytest
import torch
import torchvision.models as models
from boto3 import session
from torch.utils.data import DataLoader

import daft
from daft.catalog.__glue import GlueCatalog


@pytest.mark.integration()
def test_dataloading_from_glue_iceberg(pytestconfig):
    if pytestconfig.getoption("--credentials") is not True:
        pytest.skip("Test can only run in a credentialled environment, and when run with the `--credentials` flag")

    sess = session.Session()
    catalog = GlueCatalog.from_session(name="glue_catalog", session=sess)

    images_table = catalog.get_table("glue_iceberg_test.coco_images")
    images_df = images_table.read()  # noqa: F841

    categories_table = catalog.get_table("glue_iceberg_test.coco_categories")
    categories_df = categories_table.read().with_column_renamed("id", "category_id")  # noqa: F841

    annotations_table = catalog.get_table("glue_iceberg_test.coco_annotations")
    annotations_df = annotations_table.read().with_column_renamed("id", "annotation_id")  # noqa: F841

    df = daft.sql("""
        SELECT
            id, annotation_id, bbox, bucket_url
        FROM
            images_df
        JOIN
            annotations_df ON id = image_id
        JOIN
            categories_df USING (category_id)
        WHERE
            supercategory = 'person'
    """)
    df = df.with_column("image", df["bucket_url"].url.download().image.decode())
    df = df.with_column("cropped", df["image"].image.crop(df["bbox"]))
    df = df.with_column("thumbnail", df["cropped"].image.resize(64, 64))
    df = df.with_column("image_tensor", df["thumbnail"]).select("id", "annotation_id", "image_tensor").limit(100)

    df.collect()
    assert df.count_rows() == 100

    model = models.resnet18()
    model = torch.nn.Sequential(*list(model.children())[:-1])
    model.eval()
    device = torch.device("cpu")
    model.to(device)

    torch_dataset = df.to_torch_iter_dataset()
    dataloader = DataLoader(torch_dataset, batch_size=16)

    embeddings = []
    ids = []
    annotation_ids = []

    for batch in iter(dataloader):
        image_tensor = batch["image_tensor"].to(device)
        image_tensor = image_tensor.permute(0, 3, 1, 2).float() / 255.0
        outputs = model(image_tensor)
        outputs = outputs.view(outputs.size(0), -1)
        embeddings.append(outputs.cpu())
        ids.extend(batch["id"])
        annotation_ids.extend(batch["annotation_id"])

    all_embeddings = torch.cat(embeddings, dim=0)

    assert all_embeddings.shape[0] == len(ids), "Number of embeddings should match number of ids"
    assert all_embeddings.shape[1] == 512, "ResNet18 should produce 512-dimensional embeddings"


@pytest.mark.integration()
def test_pandas_to_daft_to_glue_iceberg(pytestconfig):
    if pytestconfig.getoption("--credentials") is not True:
        pytest.skip("Test requires AWS credentials and `--credentials` flag")

    # Create random data in pandas.
    num_rows = 10
    df_pd = pd.DataFrame(
        {
            "id": [str(uuid.uuid4()) for _ in range(num_rows)],
            "double_col": [random.random() for _ in range(num_rows)],
            "int_col": [random.randint(0, 3) for _ in range(num_rows)],
        }
    )

    # Convert to daft DataFrame.
    df = daft.from_pandas(df_pd)

    # Overwrite data in Glue Iceberg table.
    sess = session.Session()
    catalog = GlueCatalog.from_session(name="glue_catalog", session=sess)
    table = catalog.get_table("glue_iceberg_test.pandas_to_glue_table")
    table.write(df, mode="overwrite")

    # Validate that we wrote the data correctly.
    read_back = table.read()
    read_back_pd = read_back.sort("id").to_pandas().reset_index(drop=True)
    df_pd = df_pd.sort_values("id").reset_index(drop=True)
    assert read_back_pd.equals(df_pd), "Did not write data correctly to Glue Iceberg table"



from __future__ import annotations

from functools import partial
from pathlib import Path
from typing import Callable, Literal

import numpy as np
import pytest

import daft
from daft import DataType
from tests.utils import random_numerical_embedding

FMT = Literal["parquet", "lance"]


def _make_check_embeddings(
    test_df: daft.DataFrame, dtype, *, embeddings_col: str = "e"
) -> Callable[[daft.DataFrame], None]:
    """Verify validity of the test dataframe & produce a function that checks another dataframe for equality."""
    test_df = test_df.collect()
    if test_df.count_rows() == 0:
        raise ValueError("Test DataFrame cannot be empty!")

    if embeddings_col not in test_df:
        raise ValueError(f"Test DataFrame doesn't have an embeddings column={embeddings_col}")

    test_rows = test_df.to_pydict()[embeddings_col]
    for i, t in enumerate(test_rows):
        assert isinstance(t, np.ndarray), f"Row {i} is not a numpy array, it is: {type(t)}: {t}"
        assert t.dtype == dtype, f"Row {i} array doesn't have {dtype=}, instead={t.dtype}"

    def _check_embeddings(loaded_df: daft.DataFrame) -> None:
        loaded_df = loaded_df.collect()
        if loaded_df.count_rows() != test_df.count_rows():
            raise ValueError(
                f"Expecting {test_df.count_rows()} rows but got a " f"DataFrame with {loaded_df.count_rows()}"
            )

        l_rows = loaded_df.to_pydict()[embeddings_col]
        for i, (t, l) in enumerate(zip(test_rows, l_rows)):  # noqa: E741
            assert isinstance(l, np.ndarray), f"Row {i} expected a numpy array when loading, got a {type(l)}: {l}"
            assert l.dtype == t.dtype, f"Row {i} has wrong dtype. Expected={t.dtype} vs. found={l.dtype}"
            assert (t == l).all(), f"Row {i} failed equality check: test_df={t} vs. loaded={l}"

    return _check_embeddings


@pytest.mark.parametrize("fmt", ["parquet", "lance"])
@pytest.mark.parametrize(
    ["dtype", "size"],
    [
        # (np.float16, 64), -- Arrow doesn't support f16
        (np.float32, 1024),
        (np.float64, 512),
        (np.int8, 2048),
        (np.int16, 512),
        (np.int32, 256),
        (np.int64, 128),
        (np.uint8, 2048),
        (np.uint16, 512),
        (np.uint32, 256),
        (np.uint64, 128),
        # (np.bool_, 512), -- Arrow only accepts numeric types
        # (np.complex64, 32), (np.complex128, 16), - Arrow doesn't support complex numbers
    ],
)
def test_roundtrip_embedding(tmp_path: Path, fmt: FMT, dtype: np.dtype, size: int) -> None:
    # make some embeddings of the specified data type and dimensionality
    # with uniformly at random distributed values
    make_array = partial(random_numerical_embedding, np.random.default_rng(), dtype, size)
    test_df = daft.from_pydict({"e": [make_array() for _ in range(50)]}).with_column(
        "e", daft.col("e").cast(DataType.embedding(DataType.from_numpy_dtype(dtype), size))
    )

    # make a checking function for the loaded dataframe & verify our original dataframe
    check = _make_check_embeddings(test_df, dtype)

    # write the embeddings-containing dataframe to disk using the specified format
    getattr(test_df, f"write_{fmt}")(str(tmp_path))

    # read that same dataframe
    loaded_df = getattr(daft, f"read_{fmt}")(str(tmp_path))

    # check that the values in the embedding column exactly equal each other
    check(loaded_df)



from __future__ import annotations

import pytest

import daft

from .conftest import minio_create_bucket, minio_create_public_bucket


def run_url_upload_roundtrip_test(folder: str, io_config, bytes_data: list[bytes]):
    """Helper function to run URL upload/download roundtrip test."""
    data = {"data": bytes_data}
    df = daft.from_pydict(data)
    df = df.with_column("file_paths", df["data"].url.upload(folder, io_config=io_config))
    df.collect()

    df = df.with_column("roundtrip_data", df["file_paths"].url.download(io_config=io_config))
    results = df.to_pydict()

    assert results["data"] == results["roundtrip_data"] == bytes_data
    for path in results["file_paths"]:
        assert path.startswith(folder)


@pytest.mark.integration()
def test_files_roundtrip_minio_native_downloader(minio_io_config):
    bucket_name = "my-bucket"
    folder = f"s3://{bucket_name}/my-folder"
    with minio_create_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        bytes_data = [b"a", b"b", b"c"]
        run_url_upload_roundtrip_test(folder, minio_io_config, bytes_data)


@pytest.mark.integration()
def test_files_roundtrip_minio_anonymous_upload(anonymous_minio_io_config, minio_io_config):
    """Test anonymous URL upload and download roundtrip."""
    bucket_name = "my-public-bucket"
    folder = f"s3://{bucket_name}/my-folder"
    # Use the authenticated config to create the public bucket, but use the anonymous config to upload/download.
    with minio_create_public_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        bytes_data = [b"a", b"b", b"c"]
        run_url_upload_roundtrip_test(folder, anonymous_minio_io_config, bytes_data)



from __future__ import annotations

import pytest
from botocore import session

import daft
from daft.io import IOConfig, S3Config


@pytest.fixture(scope="session")
def io_config(pytestconfig) -> IOConfig:
    """Create IOConfig with botocore's current session."""
    if pytestconfig.getoption("--credentials") is not True:
        pytest.skip("Test can only run in a credentialled environment, and when run with the `--credentials` flag")

    sess = session.Session()
    creds = sess.get_credentials()

    return IOConfig(
        s3=S3Config(
            key_id=creds.access_key, access_key=creds.secret_key, session_token=creds.token, region_name="us-west-2"
        )
    )


@pytest.mark.integration()
def test_url_download_aws_s3_public_bucket_with_creds(small_images_s3_paths, io_config):
    data = {"urls": small_images_s3_paths}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(io_config=io_config))

    data = df.to_pydict()
    assert len(data["data"]) == 12
    for img_bytes in data["data"]:
        assert img_bytes is not None


@pytest.mark.integration()
def test_read_parquet_aws_s3_public_bucket_with_creds(io_config):
    filename = "s3://daft-public-data/test_fixtures/parquet-dev/mvp.parquet"
    df = daft.read_parquet(filename, io_config=io_config).collect()
    assert len(df) == 100



from __future__ import annotations

import numpy as np
import pytest

from daft.expressions import col
from tests.conftest import assert_df_equals


def test_sorted_by_expr(daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size):
    """Sort by a column that undergoes an expression."""
    daft_df = daft_df.repartition(repartition_nparts)
    daft_sorted_df = daft_df.sort(((col("Unique Key") % 2) == 0).if_else(col("Unique Key"), col("Unique Key") * -1))
    daft_sorted_pd_df = daft_sorted_df.to_pandas()

    service_requests_csv_pd_df["tmp"] = service_requests_csv_pd_df["Unique Key"]
    service_requests_csv_pd_df["tmp"] = np.where(
        service_requests_csv_pd_df["tmp"] % 2 == 0,
        service_requests_csv_pd_df["tmp"],
        service_requests_csv_pd_df["tmp"] * -1,
    )
    service_requests_csv_pd_df = service_requests_csv_pd_df.sort_values("tmp", ascending=True)
    service_requests_csv_pd_df = service_requests_csv_pd_df.drop(["tmp"], axis=1)

    assert_df_equals(
        daft_sorted_pd_df,
        service_requests_csv_pd_df,
        assert_ordering=True,
    )


@pytest.mark.parametrize(
    "sort_keys",
    [
        pytest.param(["Unique Key"], id="NumSortKeys:1"),
        pytest.param(["Borough", "Unique Key"], id="NumSortKeys:2"),
    ],
)
def test_get_sorted(daft_df, service_requests_csv_pd_df, repartition_nparts, sort_keys, with_morsel_size):
    """Sort by a column."""
    daft_df = daft_df.repartition(repartition_nparts)
    daft_sorted_df = daft_df.sort([col(k) for k in sort_keys], desc=True)
    daft_sorted_pd_df = daft_sorted_df.to_pandas()
    assert_df_equals(
        daft_sorted_pd_df,
        service_requests_csv_pd_df.sort_values(by=sort_keys, ascending=False),
        assert_ordering=True,
    )


@pytest.mark.parametrize(
    "sort_keys",
    [
        pytest.param(["Unique Key"], id="NumSortKeys:1"),
        pytest.param(["Borough", "Unique Key"], id="NumSortKeys:2"),
    ],
)
def test_get_sorted_top_n(daft_df, service_requests_csv_pd_df, repartition_nparts, sort_keys, with_morsel_size):
    """Sort by a column."""
    daft_df = daft_df.repartition(repartition_nparts)
    daft_sorted_df = daft_df.sort([col(k) for k in sort_keys], desc=True).limit(100)
    daft_sorted_pd_df = daft_sorted_df.to_pandas()

    assert_df_equals(
        daft_sorted_pd_df,
        service_requests_csv_pd_df.sort_values(by=sort_keys, ascending=False).head(100),
        assert_ordering=True,
    )


@pytest.mark.parametrize(
    "sort_keys",
    [
        pytest.param(["Borough", "Unique Key"], id="NumSortKeys:2"),
    ],
)
def test_get_sorted_top_n_flipped_desc(
    daft_df, service_requests_csv_pd_df, repartition_nparts, sort_keys, with_morsel_size
):
    """Sort by a column."""
    daft_df = daft_df.repartition(repartition_nparts)
    desc_list = [True]
    for i in range(len(sort_keys) - 1):
        desc_list.append(not desc_list[-1])
    daft_sorted_df = daft_df.sort([col(k) for k in sort_keys], desc=desc_list).limit(100)
    daft_sorted_pd_df = daft_sorted_df.to_pandas()

    assert_df_equals(
        daft_sorted_pd_df,
        service_requests_csv_pd_df.sort_values(by=sort_keys, ascending=[not b for b in desc_list]).head(100),
        assert_ordering=True,
    )


@pytest.mark.parametrize(
    "daft_df_ops",
    [
        pytest.param(
            lambda daft_df: daft_df.sort(col("Unique Key"), desc=True).select(col("Unique Key"), col("Complaint Type")),
            id="sort..select",
        ),
        pytest.param(
            lambda daft_df: daft_df.select(col("Unique Key"), col("Complaint Type")).sort(col("Unique Key"), desc=True),
            id="select..sort",
        ),
    ],
)
def test_get_sorted_top_n_projected(
    daft_df_ops, daft_df, service_requests_csv_pd_df, repartition_nparts, with_morsel_size
):
    """Sort by a column and retrieve specific columns from the top N results."""
    daft_df = daft_df.repartition(repartition_nparts)
    expected = service_requests_csv_pd_df.sort_values(by="Unique Key", ascending=False)[
        ["Unique Key", "Complaint Type"]
    ]
    daft_df = daft_df_ops(daft_df)
    daft_pd_pf = daft_df.to_pandas()
    assert_df_equals(
        daft_pd_pf,
        expected,
        assert_ordering=True,
    )



from __future__ import annotations

import sys

import pytest

import daft
from benchmarking.tpch import answers

if sys.platform == "win32":
    pytest.skip(allow_module_level=True)

from tests.conftest import get_tests_daft_runner_name
from tests.integration.conftest import check_answer  # noqa F401

TPCH_QUESTIONS = list(range(1, 11))


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native",
    reason="requires Native Runner to be in use",
)
@pytest.mark.benchmark(group="tpch")
@pytest.mark.parametrize("q", TPCH_QUESTIONS)
def test_tpch(tmp_path, check_answer, get_df, benchmark_with_memray, q):  # noqa F811
    get_df, num_parts = get_df

    def f():
        question = getattr(answers, f"q{q}")
        daft_df = question(get_df)
        return daft_df.to_arrow()

    benchmark_group = f"q{q}-parts-{num_parts}"
    daft_pd_df = benchmark_with_memray(f, benchmark_group).to_pandas()
    check_answer(daft_pd_df, q, tmp_path)


@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native",
    reason="requires Native Runner to be in use",
)
@pytest.mark.benchmark(group="tpch")
@pytest.mark.parametrize("q", TPCH_QUESTIONS)
def test_tpch_sql(tmp_path, check_answer, get_df, benchmark_with_memray, q):  # noqa F811
    from daft.sql import SQLCatalog

    get_df, num_parts = get_df

    # TODO: remove this once SQL allows case-insensitive column names
    def lowercase_column_names(df):
        return df.select(*[daft.col(name).alias(name.lower()) for name in df.column_names])

    table_names = [
        "part",
        "supplier",
        "partsupp",
        "customer",
        "orders",
        "lineitem",
        "nation",
        "region",
    ]
    catalog = SQLCatalog({tbl: lowercase_column_names(get_df(tbl)) for tbl in table_names})

    with open(f"benchmarking/tpch/queries/{q:02}.sql") as query_file:
        query = query_file.read()

    def f():
        daft_df = daft.sql(query, catalog=catalog)
        return daft_df.to_arrow()

    benchmark_group = f"q{q}-sql-parts-{num_parts}"
    daft_pd_df = benchmark_with_memray(f, benchmark_group).to_pandas()
    check_answer(daft_pd_df, q, tmp_path)



# ruff: noqa: I002
# isort: dont-add-import: from __future__ import annotations
from typing import Optional, Union

from daft import context
from daft.api_annotations import PublicAPI
from daft.daft import IOConfig, ScanOperatorHandle
from daft.dataframe import DataFrame
from daft.io.lance.lance_scan import LanceDBScanOperator
from daft.io.object_store_options import io_config_to_storage_options
from daft.logical.builder import LogicalPlanBuilder


@PublicAPI
def read_lance(
    url: str,
    io_config: Optional[IOConfig] = None,
    version: Optional[Union[str, int]] = None,
    asof: Optional[str] = None,
    block_size: Optional[int] = None,
    commit_lock: Optional[object] = None,
    index_cache_size: Optional[int] = None,
    default_scan_options: Optional[dict[str, str]] = None,
    metadata_cache_size_bytes: Optional[int] = None,
) -> DataFrame:
    """Create a DataFrame from a LanceDB table.

    Args:
        url: URL to the LanceDB table (supports remote URLs to object stores such as `s3://` or `gs://`)
        io_config: A custom IOConfig to use when accessing LanceDB data. Defaults to None.
        version : optional, int | str
            If specified, load a specific version of the Lance dataset. Else, loads the
            latest version. A version number (`int`) or a tag (`str`) can be provided.
        asof : optional, datetime or str
            If specified, find the latest version created on or earlier than the given
            argument value. If a version is already specified, this arg is ignored.
        block_size : optional, int
            Block size in bytes. Provide a hint for the size of the minimal I/O request.
        commit_lock : optional, lance.commit.CommitLock
            A custom commit lock.  Only needed if your object store does not support
            atomic commits.  See the user guide for more details.
        index_cache_size : optional, int
            Index cache size. Index cache is a LRU cache with TTL. This number specifies the
            number of index pages, for example, IVF partitions, to be cached in
            the host memory. Default value is ``256``.

            Roughly, for an ``IVF_PQ`` partition with ``n`` rows, the size of each index
            page equals the combination of the pq code (``np.array([n,pq], dtype=uint8))``
            and the row ids (``np.array([n], dtype=uint64)``).
            Approximately, ``n = Total Rows / number of IVF partitions``.
            ``pq = number of PQ sub-vectors``.
        storage_options : optional, dict
            Extra options that make sense for a particular storage connection. This is
            used to store connection parameters like credentials, endpoint, etc.
        default_scan_options : optional, dict
            Default scan options that are used when scanning the dataset.  This accepts
            the same arguments described in :py:meth:`lance.LanceDataset.scanner`.  The
            arguments will be applied to any scan operation.

            This can be useful to supply defaults for common parameters such as
            ``batch_size``.

            It can also be used to create a view of the dataset that includes meta
            fields such as ``_rowid`` or ``_rowaddr``.  If ``default_scan_options`` is
            provided then the schema returned by :py:meth:`lance.LanceDataset.schema` will
            include these fields if the appropriate scan options are set.
        metadata_cache_size_bytes : optional, int
            Size of the metadata cache in bytes. This cache is used to store metadata
            information about the dataset, such as schema and statistics. If not specified,
            a default size will be used.

    Returns:
        DataFrame: a DataFrame with the schema converted from the specified LanceDB table

    Note:
        This function requires the use of [LanceDB](https://lancedb.github.io/lancedb/), which is the Python library for the LanceDB project.
        To ensure that this is installed with Daft, you may install: `pip install daft[lance]`

    Examples:
        Read a local LanceDB table:
        >>> df = daft.read_lance("s3://my-lancedb-bucket/data/")
        >>> df.show()

        Read a LanceDB table from a public S3 bucket:
        >>> from daft.io import S3Config
        >>> s3_config = S3Config(region="us-west-2", anonymous=True)
        >>> df = daft.read_lance("s3://daft-public-data/lance/words-test-dataset", io_config=s3_config)
        >>> df.show()

        Read a local LanceDB table and specify a version:
        >>> df = daft.read_lance("s3://my-lancedb-bucket/data/", version=1)
        >>> df.show()
    """
    try:
        import lance
    except ImportError as e:
        raise ImportError(
            "Unable to import the `lance` package, please ensure that Daft is installed with the lance extra dependency: `pip install daft[lance]`"
        ) from e

    io_config = context.get_context().daft_planning_config.default_io_config if io_config is None else io_config
    storage_options = io_config_to_storage_options(io_config, url)

    ds = lance.dataset(
        url,
        storage_options=storage_options,
        version=version,
        asof=asof,
        block_size=block_size,
        commit_lock=commit_lock,
        index_cache_size=index_cache_size,
        default_scan_options=default_scan_options,
        metadata_cache_size_bytes=metadata_cache_size_bytes,
    )
    lance_operator = LanceDBScanOperator(ds)

    handle = ScanOperatorHandle.from_python_scan_operator(lance_operator)
    builder = LogicalPlanBuilder.from_tabular_scan(scan_operator=handle)
    return DataFrame(builder)



from __future__ import annotations

from collections.abc import Iterator

import pytest
import ray

import daft
from daft.io.sink import DataSink, WriteResult
from daft.recordbatch import MicroPartition


class UnserializableException(Exception):
    def __init__(self, message="Recovered the message in the exception"):
        super().__init__(message)
        self.unsafe_lambda = lambda x: x * 2


class SampleSink(DataSink[None]):
    def name(self) -> str:
        return "sample_sink"

    def schema(self) -> daft.Schema:
        return daft.Schema.from_pydict({"id": daft.DataType.int64()})

    def write(self, df: daft.DataFrame) -> Iterator[WriteResult[None]]:
        raise UnserializableException()

    def finalize(self, write_results: list[WriteResult[None]]) -> MicroPartition:
        result_table = MicroPartition.from_pydict(
            {
                "write_responses": write_results,
            }
        )

        return result_table


def test_sink_raises_unserializable_exception():
    df = daft.from_pydict({"id": [1, 2, 3]})
    sink = SampleSink()
    with pytest.raises(Exception) as exc_info:
        df.write_sink(sink)

    e = exc_info.value
    assert isinstance(e, (RuntimeError, ray.exceptions.RayTaskError))
    assert "UnserializableException" in str(e)



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.series import Series
from daft.utils import pyarrow_supports_fixed_shape_tensor
from tests.conftest import get_tests_daft_runner_name
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES, ARROW_STRING_TYPES

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES + ARROW_STRING_TYPES)
def test_series_take(dtype) -> None:
    data = pa.array([1, 2, 3, None, 5, None])

    s = Series.from_arrow(data.cast(dtype))
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


def test_series_date_take() -> None:
    from datetime import date

    def date_maker(d):
        if d is None:
            return None
        return date(2023, 1, d)

    days = list(map(date_maker, [5, 4, 1, None, 2, None]))
    s = Series.from_pylist(days)
    taken = s.take(Series.from_pylist([5, 4, 3, 2, 1, 0]))
    assert taken.datatype() == DataType.date()
    assert taken.to_pylist() == days[::-1]


@pytest.mark.parametrize("time_unit", ["us", "ns"])
def test_series_time_take(time_unit) -> None:
    from datetime import time

    def time_maker(h, m, s, us):
        if us is None:
            return None
        return time(h, m, s, us)

    times = list(map(time_maker, [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [0, 0, 0, 0, 0, 0], [5, 4, 1, None, 2, None]))
    s = Series.from_pylist(times)
    s = s.cast(DataType.time(time_unit))
    taken = s.take(Series.from_pylist([5, 4, 3, 2, 1, 0]))
    assert taken.datatype() == DataType.time(time_unit)
    assert taken.to_pylist() == times[::-1]


@pytest.mark.parametrize("type", [pa.binary(), pa.binary(1)])
def test_series_binary_take(type) -> None:
    data = pa.array([b"1", b"2", b"3", None, b"5", None], type=type)

    s = Series.from_arrow(data)
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


def test_series_list_take() -> None:
    data = pa.array([[1, 2], [None], None, [7, 8, 9], [10, None], [11]], type=pa.list_(pa.int64()))

    s = Series.from_arrow(data)
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


def test_series_fixed_size_list_take() -> None:
    data = pa.array([[1, 2], [None, 4], None, [7, 8], [9, None], [11, 12]], type=pa.list_(pa.int64(), 2))

    s = Series.from_arrow(data)
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


def test_series_struct_take() -> None:
    dtype = pa.struct({"a": pa.int64(), "b": pa.float64(), "c": pa.string()})
    data = pa.array(
        [
            {"a": 1, "b": 2},
            {"b": 3, "c": "4"},
            None,
            {"a": 5, "b": 6, "c": "7"},
            {"a": 8, "b": None, "c": "10"},
            {"b": 11, "c": None},
        ],
        type=dtype,
    )

    s = Series.from_arrow(data)
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


@pytest.mark.skipif(
    get_tests_daft_runner_name() == "ray",
    reason="pyarrow extension types aren't supported on Ray clusters.",
)
def test_series_extension_type_take(uuid_ext_type) -> None:
    pydata = [f"{i}".encode() for i in range(6)]
    pydata[2] = None
    storage = pa.array(pydata)
    data = pa.ExtensionArray.from_storage(uuid_ext_type, storage)

    s = Series.from_arrow(data)
    assert s.datatype() == DataType.extension(
        uuid_ext_type.NAME, DataType.from_arrow_type(uuid_ext_type.storage_type), ""
    )
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    expected = [pydata[i] if i is not None else None for i in pyidx]
    assert result.to_pylist() == expected


@pytest.mark.skipif(
    not pyarrow_supports_fixed_shape_tensor(),
    reason=f"Arrow version {ARROW_VERSION} doesn't support the canonical tensor extension type.",
)
def test_series_canonical_tensor_extension_type_take() -> None:
    pydata = np.arange(24).reshape((6, 4)).tolist()
    pydata[2] = None
    storage = pa.array(pydata, pa.list_(pa.int64(), 4))
    shape = (2, 2)
    tensor_type = pa.fixed_shape_tensor(pa.int64(), shape)
    data = pa.FixedShapeTensorArray.from_storage(tensor_type, storage)

    s = Series.from_arrow(data)
    assert s.datatype() == DataType.tensor(DataType.from_arrow_type(tensor_type.storage_type.value_type), shape)
    pyidx = [2, 0, None, 5]
    idx = Series.from_pylist(pyidx)

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 4

    original_data = s.to_pylist()
    expected = [original_data[i] if i is not None else None for i in pyidx]
    np.testing.assert_equal(result.to_pylist(), expected)


def test_series_deeply_nested_take() -> None:
    # Test take on a Series with a deeply nested type: struct of list of struct of list of strings.
    data = pa.array([{"a": [{"b": ["foo", "bar"]}]}, {"a": [{"b": ["baz", "quux"]}]}])
    dtype = pa.struct([("a", pa.large_list(pa.struct([("b", pa.large_list(pa.large_string()))])))])

    s = Series.from_arrow(data)
    assert s.datatype() == DataType.from_arrow_type(dtype)
    idx = Series.from_pylist([1])

    result = s.take(idx)
    assert result.datatype() == s.datatype()
    assert len(result) == 1

    original_data = s.to_pylist()
    expected = [original_data[1]]
    assert result.to_pylist() == expected



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Callable, Literal, TypeVar

import daft.daft as native
from daft.arrow_utils import ensure_array, ensure_chunked_array
from daft.daft import CountMode, ImageFormat, ImageMode, PyExpr, PyRecordBatch, PySeries, series_lit
from daft.datatype import DataType, TimeUnit, _ensure_registered_super_ext_type
from daft.dependencies import np, pa, pd
from daft.utils import pyarrow_supports_fixed_shape_tensor

if TYPE_CHECKING:
    import builtins
    from collections.abc import Iterator

    from daft.daft import PyDataType


class SeriesIterable:
    """Iterable wrapper for Series that efficiently handles different data types."""

    def __init__(self, series: Series):
        self.series = series

    def __iter__(self) -> Iterator[Any]:
        dt = self.series.datatype()
        if dt == DataType.python():

            def yield_pylist() -> Iterator[Any]:
                yield from self.series._series.to_pylist()

            return yield_pylist()
        elif dt._should_cast_to_python():

            def yield_pylist() -> Iterator[Any]:
                yield from self.series._series.cast(DataType.python()._dtype).to_pylist()

            return yield_pylist()
        else:

            def arrow_to_py() -> Iterator[Any]:
                # We directly call .to_arrow() on the internal PySeries object since the case
                # above has already captured the fixed shape tensor case.
                arrow_data = self.series._series.to_arrow()
                for item in arrow_data:
                    yield None if item is None else item.as_py()

            return arrow_to_py()


class Series:
    """A Daft Series is an array of data of a single type, and is usually a column in a DataFrame."""

    _series: PySeries

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a Series via __init__ ")

    def __iter__(self) -> Iterator[Any]:
        """Return an iterator over the elements of the Series."""
        iterable = SeriesIterable(self)
        return iterable.__iter__()

    @staticmethod
    def _from_pyseries(pyseries: PySeries) -> Series:
        s = Series.__new__(Series)
        s._series = pyseries
        return s

    @staticmethod
    def from_arrow(
        array: pa.Array | pa.ChunkedArray, name: str = "arrow_series", dtype: DataType | None = None
    ) -> Series:
        """Construct a Series from an pyarrow array or chunked array.

        Args:
            array: The pyarrow (chunked) array whose data we wish to put in the Series.
            name: The name associated with the Series; this is usually the column name.
            dtype: The DataType to use for the Series. If not provided, Daft will infer the
                DataType from the data.
        """
        _ensure_registered_super_ext_type()
        if (dtype and dtype.is_python()) or DataType.from_arrow_type(array.type).is_python():
            # If the Arrow type is not natively supported, go through the Python list path.
            return Series.from_pylist(array.to_pylist(), name=name, pyobj="force")
        elif isinstance(array, pa.Array):
            array = ensure_array(array)
            if isinstance(array.type, getattr(pa, "FixedShapeTensorType", ())):
                series = Series.from_arrow(array.storage, name=name)
                return series.cast(dtype or DataType.from_arrow_type(array.type))
            else:
                pys = PySeries.from_arrow(name, array, dtype=dtype._dtype if dtype else None)
                return Series._from_pyseries(pys)
        elif isinstance(array, pa.ChunkedArray):
            array = ensure_chunked_array(array)
            arr_type = array.type
            if isinstance(arr_type, pa.BaseExtensionType):
                combined_storage_array = array.cast(arr_type.storage_type).combine_chunks()
                combined_array = arr_type.wrap_array(combined_storage_array)
            else:
                combined_array = array.combine_chunks()
            return Series.from_arrow(combined_array, name=name, dtype=dtype)
        else:
            raise TypeError(f"expected either PyArrow Array or Chunked Array, got {type(array)}")

    @staticmethod
    def from_pylist(
        data: list[Any],
        name: str = "list_series",
        dtype: DataType | None = None,
        pyobj: Literal["allow", "disallow", "force"] = "allow",
    ) -> Series:
        """Construct a Series from a Python list.

        If `dtype` is not defined, then the resulting type depends on the setting of `pyobj`:
            - ``"allow"``: Arrow-backed types if possible, else PyObject;
            - ``"disallow"``: Arrow-backed types only, raising error if not convertible;
            - ``"force"``: Always store as PyObject types. Equivalent to `dtype=daft.DataType.python()`.

        Args:
            data: The Python list whose data we wish to put in the Series.
            name: The name associated with the Series; this is usually the column name.
            dtype: The DataType to use for the Series. If not provided, Daft will infer the
                DataType from the data.
            pyobj: Whether we want to ``"allow"`` coercion to Arrow types, ``"disallow"``
                falling back to Python type representation. Default is ``"allow"``.
        """
        if not isinstance(data, list):
            raise TypeError(f"expected a python list, got {type(data)}")

        if pyobj not in {"allow", "disallow", "force"}:
            raise ValueError(f"pyobj: expected either 'allow', 'disallow', or 'force', but got {pyobj})")

        # If output is Python objects, we can just use the Python list directly.
        if (dtype and dtype == DataType.python()) or pyobj == "force":
            pys = PySeries.from_pylist(name, data, DataType.python()._dtype)
            return Series._from_pyseries(pys)

        # Otherwise, try to infer from parameters provided
        try:
            # Workaround: wrap list of np.datetime64 in an np.array
            #   - https://github.com/apache/arrow/issues/40580
            #   - https://github.com/Eventual-Inc/Daft/issues/3826
            if data and np.module_available() and isinstance(data[0], np.datetime64):  # type: ignore[attr-defined]
                np_arr = np.array(data)
                arrow_array = pa.array(np_arr)
            else:
                arrow_array = pa.array(data, type=dtype.to_arrow_dtype() if dtype else None)
            return Series.from_arrow(arrow_array, name=name, dtype=dtype)
        except (pa.lib.ArrowInvalid, pa.lib.ArrowTypeError, pa.lib.ArrowNotImplementedError):
            if pyobj == "disallow":
                raise
            dtype = dtype or DataType._infer_dtype_from_pylist(data) or DataType.python()
            pys = PySeries.from_pylist(name, data, dtype._dtype)
            return Series._from_pyseries(pys)

    @classmethod
    def from_numpy(
        cls, data: np.ndarray[Any, Any], name: str = "numpy_series", dtype: DataType | None = None
    ) -> Series:
        """Construct a Series from a NumPy ndarray.

        If the provided NumPy ndarray is 1-dimensional, Daft will attempt to store the ndarray
        in a pyarrow Array. If the ndarray has more than 1 dimension OR storing the 1D array in Arrow failed,
        Daft will store the ndarray data as a Python list of NumPy ndarrays.

        Args:
            data: The NumPy ndarray whose data we wish to put in the Series.
            name: The name associated with the Series; this is usually the column name.
            dtype: The DataType to use for the Series. If not provided, Daft will infer the
                DataType from the data.
        """
        if not isinstance(data, np.ndarray):
            raise TypeError(f"Expected a NumPy ndarray, got {type(data)}")
        if data.ndim <= 1:
            try:
                arrow_array = pa.array(data)
            except pa.ArrowInvalid:
                pass
            else:
                return cls.from_arrow(arrow_array, name=name, dtype=dtype)
        # TODO(Clark): Represent the tensor series with an Arrow extension type in order
        # to keep the series data contiguous.
        list_ndarray = [np.asarray(item) for item in data]
        return cls.from_pylist(list_ndarray, name=name, dtype=dtype)

    @classmethod
    def from_pandas(cls, data: pd.Series[Any], name: str = "pd_series", dtype: DataType | None = None) -> Series:
        """Construct a Series from a pandas Series.

        This will first try to convert the series into a pyarrow array, then will fall
        back to converting the series to a NumPy ndarray and going through that construction path,
        and will finally fall back to converting the series to a Python list and going through that
        path.

        Args:
            data: The pandas Series whose data we wish to put in the Daft Series.
            name: The name associated with the Series; this is usually the column name.
            dtype: The DataType to use for the Series. If not provided, Daft will infer the
                DataType from the data.
        """
        if not isinstance(data, pd.Series):
            raise TypeError(f"expected a pandas Series, got {type(data)}")
        # First, try Arrow path.
        try:
            arrow_arr = pa.Array.from_pandas(data)
        except pa.ArrowInvalid:
            pass
        else:
            return cls.from_arrow(arrow_arr, name=name, dtype=dtype)
        # Second, fall back to NumPy path. Note that .from_numpy() does _not_ fall back to
        # the pylist representation for 1D arrays and instead raises an error that we can catch.
        # We do the pylist representation fallback ourselves since the pd.Series.to_list()
        # preserves more type information for types that are not natively representable in Python.
        try:
            ndarray = data.to_numpy()
            return cls.from_numpy(ndarray, name=name, dtype=dtype)
        except Exception:
            pass
        # Finally, fall back to pylist path.
        # NOTE: For element types that don't have a native Python representation,
        # a Pandas scalar object will be returned.
        return cls.from_pylist(data.to_list(), name=name, dtype=dtype if dtype else DataType.python())

    def cast(self, dtype: DataType) -> Series:
        return Series._from_pyseries(self._series.cast(dtype._dtype))

    def _cast_to_python(self) -> Series:
        """Convert this Series into a Series of Python objects.

        Call Series.to_pylist() and create a new Series from the raw Pylist directly.

        This logic is needed by the Rust implementation of cast(),
        but is written here (i.e. not in Rust) for conciseness.

        Do not call this method directly in Python; call cast() instead.
        """
        pylist = self.to_pylist()
        return Series.from_pylist(pylist, self.name(), dtype=DataType.python())

    def _pycast_to_pynative(self, typefn: type, dtype: PyDataType) -> Series:
        """Apply Python-level casting to this Series.

        Call Series.to_pylist(), apply the Python cast (e.g. str(x)),
        and create a new arrow-backed Series from the result.

        This logic is needed by the Rust implementation of cast(),
        but is written here (i.e. not in Rust) for conciseness.

        Do not call this method directly in Python; call cast() instead.
        """
        pylist = self.to_pylist()
        pylist = [typefn(val) if val is not None else None for val in pylist]
        return Series.from_pylist(pylist, self.name(), pyobj="disallow", dtype=DataType._from_pydatatype(dtype))

    @staticmethod
    def concat(series: list[Series]) -> Series:
        pyseries = []
        for s in series:
            if not isinstance(s, Series):
                raise TypeError(f"Expected a Series for concat, got {type(s)}")
            pyseries.append(s._series)
        return Series._from_pyseries(PySeries.concat(pyseries))

    def name(self) -> str:
        return self._series.name()

    def rename(self, name: str) -> Series:
        return Series._from_pyseries(self._series.rename(name))

    def datatype(self) -> DataType:
        return DataType._from_pydatatype(self._series.data_type())

    def to_arrow(self) -> pa.Array:
        """Convert this Series to an pyarrow array."""
        _ensure_registered_super_ext_type()

        dtype = self.datatype()

        # Special-case for PyArrow FixedShapeTensor if it is supported by the version of PyArrow
        # TODO: Push this down into self._series.to_arrow()?
        if dtype.is_fixed_shape_tensor() and pyarrow_supports_fixed_shape_tensor():
            pyarrow_dtype = dtype.to_arrow_dtype()
            arrow_series = self._series.to_arrow()
            return pa.ExtensionArray.from_storage(pyarrow_dtype, arrow_series.storage)
        else:
            return self._series.to_arrow()

    def to_pylist(self) -> list[Any]:
        """Convert this Series to a Python list."""
        if self.datatype().is_python():
            return self._series.to_pylist()
        elif self.datatype()._should_cast_to_python():
            return self._series.cast(DataType.python()._dtype).to_pylist()
        else:
            return self._series.to_arrow().to_pylist()

    def filter(self, mask: Series) -> Series:
        if not isinstance(mask, Series):
            raise TypeError(f"expected another Series but got {type(mask)}")
        return Series._from_pyseries(self._series.filter(mask._series))

    def take(self, idx: Series) -> Series:
        if not isinstance(idx, Series):
            raise TypeError(f"expected another Series but got {type(idx)}")
        return Series._from_pyseries(self._series.take(idx._series))

    def slice(self, start: int, end: int) -> Series:
        if not isinstance(start, int):
            raise TypeError(f"expected int for start but got {type(start)}")
        if not isinstance(end, int):
            raise TypeError(f"expected int for end but got {type(end)}")

        return Series._from_pyseries(self._series.slice(start, end))

    def argsort(self, descending: bool = False, nulls_first: bool | None = None) -> Series:
        if not isinstance(descending, bool):
            raise TypeError(f"expected `descending` to be bool, got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending

        return Series._from_pyseries(self._series.argsort(descending, nulls_first))

    def sort(self, descending: bool = False, nulls_first: bool | None = None) -> Series:
        if not isinstance(descending, bool):
            raise TypeError(f"expected `descending` to be bool, got {type(descending)}")
        if nulls_first is None:
            nulls_first = descending
        return Series._from_pyseries(self._series.sort(descending, nulls_first))

    def hash(self, seed: Series | None = None) -> Series:
        return self._eval_expressions("hash", seed=seed)

    def murmur3_32(self) -> Series:
        return Series._from_pyseries(self._series.murmur3_32())

    def __repr__(self) -> str:
        return repr(self._series)

    def __bool__(self) -> bool:
        raise ValueError(
            "Series don't have a truth value." "If you reached this error using `and` / `or`, use `&` / `|` instead."
        )

    def __len__(self) -> int:
        return len(self._series)

    def size_bytes(self) -> int:
        """Returns the total sizes of all buffers used for representing this Series.

        In particular, this includes the:

        1. Buffer(s) used for data (applies any slicing if that occurs!)
        2. Buffer(s) used for offsets, if applicable (for variable-length arrow types)
        3. Buffer(s) used for validity, if applicable (arrow can choose to omit the validity bitmask)
        4. Recursively gets .size_bytes for any child arrays, if applicable (for nested types)
        """
        return self._series.size_bytes()

    def __abs__(self) -> Series:
        return Series._from_pyseries(abs(self._series))

    def ceil(self) -> Series:
        return self._eval_expressions("ceil")

    def floor(self) -> Series:
        return self._eval_expressions("floor")

    def sign(self) -> Series:
        """The sign of a numeric series."""
        return self._eval_expressions("sign")

    def signum(self) -> Series:
        """The signum of a numeric series."""
        return self._eval_expressions("sign")

    def negate(self) -> Series:
        """The negative of a numeric series."""
        return self._eval_expressions("negative")

    def negative(self) -> Series:
        """The negative of a numeric series."""
        return self._eval_expressions("negative")

    def round(self, decimals: int = 0) -> Series:
        return self._eval_expressions("round", decimals=decimals)

    def clip(self, min: Series, max: Series) -> Series:
        return self._eval_expressions("clip", min, max)

    def sqrt(self) -> Series:
        return self._eval_expressions("sqrt")

    def cbrt(self) -> Series:
        return self._eval_expressions("cbrt")

    def sin(self) -> Series:
        """The elementwise sine of a numeric series."""
        return self._eval_expressions("sin")

    def cos(self) -> Series:
        """The elementwise cosine of a numeric series."""
        return self._eval_expressions("cos")

    def tan(self) -> Series:
        """The elementwise tangent of a numeric series."""
        return self._eval_expressions("tan")

    def csc(self) -> Series:
        """The elementwise cosecant of a numeric series."""
        return self._eval_expressions("csc")

    def sec(self) -> Series:
        """The elementwise secant of a numeric series."""
        return self._eval_expressions("sec")

    def cot(self) -> Series:
        """The elementwise cotangent of a numeric series."""
        return self._eval_expressions("cot")

    def sinh(self) -> Series:
        """The elementwise hyperbolic sine of a numeric series."""
        return self._eval_expressions("sinh")

    def cosh(self) -> Series:
        """The elementwise hyperbolic cosine of a numeric series."""
        return self._eval_expressions("cosh")

    def tanh(self) -> Series:
        """The elementwise hyperbolic tangent of a numeric series."""
        return self._eval_expressions("tanh")

    def arcsin(self) -> Series:
        """The elementwise arc sine of a numeric series."""
        return self._eval_expressions("arcsin")

    def arccos(self) -> Series:
        """The elementwise arc cosine of a numeric series."""
        return self._eval_expressions("arccos")

    def arctan(self) -> Series:
        """The elementwise arc tangent of a numeric series."""
        return self._eval_expressions("arctan")

    def arctan2(self, other: Series) -> Series:
        """Calculates the four quadrant arctangent of coordinates (y, x)."""
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        return self._eval_expressions("arctan2", other)

    def arctanh(self) -> Series:
        """The elementwise inverse hyperbolic tangent of a numeric series."""
        return self._eval_expressions("arctanh")

    def arccosh(self) -> Series:
        """The elementwise inverse hyperbolic cosine of a numeric series."""
        return self._eval_expressions("arccosh")

    def arcsinh(self) -> Series:
        """The elementwise inverse hyperbolic sine of a numeric series."""
        return self._eval_expressions("arcsinh")

    def radians(self) -> Series:
        """The elementwise radians of a numeric series."""
        return self._eval_expressions("radians")

    def degrees(self) -> Series:
        """The elementwise degrees of a numeric series."""
        return self._eval_expressions("degrees")

    def log2(self) -> Series:
        """The elementwise log2 of a numeric series."""
        return self._eval_expressions("log2")

    def log10(self) -> Series:
        """The elementwise log10 of a numeric series."""
        return self._eval_expressions("log10")

    def log(self, base: float) -> Series:
        """The elementwise log with given base, of a numeric series.

        Args:
            base: The base of the logarithm.
        """
        return self._eval_expressions("log", base=base)

    def ln(self) -> Series:
        """The elementwise ln of a numeric series."""
        return self._eval_expressions("ln")

    def log1p(self) -> Series:
        """The ln(self + 1) of a numeric series."""
        return self._eval_expressions("log1p")

    def exp(self) -> Series:
        """The e^self of a numeric series."""
        return self._eval_expressions("exp")

    def expm1(self) -> Series:
        """The e^self - 1 of a numeric series."""
        return self._eval_expressions("expm1")

    def __add__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series + other._series)

    def __sub__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series - other._series)

    def __mul__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series * other._series)

    def __truediv__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series / other._series)

    def __mod__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series % other._series)

    def __eq__(self, other: object) -> Series:  # type: ignore[override]
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series == other._series)

    def __ne__(self, other: object) -> Series:  # type: ignore[override]
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series != other._series)

    def __gt__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series > other._series)

    def __lt__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series < other._series)

    def __ge__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series >= other._series)

    def __le__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series <= other._series)

    def __lshift__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series << other._series)

    def __rshift__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series >> other._series)

    def __invert__(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.__invert__())

    def __and__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series & other._series)

    def __or__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series | other._series)

    def __xor__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series ^ other._series)

    def __floordiv__(self, other: object) -> Series:
        if not isinstance(other, Series):
            raise TypeError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series // other._series)

    def count(self, mode: CountMode = CountMode.Valid) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.count(mode))

    def min(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.min())

    def max(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.max())

    def mean(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.mean())

    def stddev(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.stddev())

    def sum(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.sum())

    def shift_right(self, bits: Series) -> Series:
        if not isinstance(bits, Series):
            raise TypeError(f"expected another Series but got {type(bits)}")
        assert self._series is not None and bits._series is not None
        return Series._from_pyseries(self._series >> bits._series)

    def shift_left(self, bits: Series) -> Series:
        if not isinstance(bits, Series):
            raise TypeError(f"expected another Series but got {type(bits)}")
        assert self._series is not None and bits._series is not None
        return Series._from_pyseries(self._series << bits._series)

    def if_else(self, if_true: object, if_false: object) -> Series:
        if not isinstance(if_true, Series):
            raise ValueError(f"expected another Series but got {type(if_true)}")
        if not isinstance(if_false, Series):
            raise ValueError(f"expected another Series but got {type(if_false)}")
        assert self._series is not None and if_true._series is not None and if_false._series is not None
        # NOTE: Rust Series has a different ordering for if_else because of better static typing
        return Series._from_pyseries(if_true._series.if_else(if_false._series, self._series))

    def is_null(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.is_null())

    def not_null(self) -> Series:
        assert self._series is not None
        return Series._from_pyseries(self._series.not_null())

    def fill_null(self, fill_value: object) -> Series:
        if not isinstance(fill_value, Series):
            raise ValueError(f"expected another Series but got {type(fill_value)}")
        assert self._series is not None and fill_value._series is not None
        return Series._from_pyseries(self._series.fill_null(fill_value._series))

    def minhash(
        self,
        num_hashes: int,
        ngram_size: int,
        seed: int = 1,
        hash_function: Literal["murmurhash3", "xxhash", "sha1"] = "murmurhash3",
    ) -> Series:
        """Runs the MinHash algorithm on the series.

        For a string, calculates the minimum hash over all its ngrams,
        repeating with `num_hashes` permutations. Returns as a list of 32-bit unsigned integers.

        Tokens for the ngrams are delimited by spaces.
        MurmurHash is used for the initial hash.

        Args:
            num_hashes: The number of hash permutations to compute.
            ngram_size: The number of tokens in each shingle/ngram.
            seed (optional): Seed used for generating permutations and the initial string hashes. Defaults to 1.
            hash_function (optional): Hash function to use for initial string hashing. One of "murmur3", "xxhash", or "sha1". Defaults to "murmur3".
        """
        if not isinstance(num_hashes, int):
            raise ValueError(f"expected an integer for num_hashes but got {type(num_hashes)}")
        if not isinstance(ngram_size, int):
            raise ValueError(f"expected an integer for ngram_size but got {type(ngram_size)}")
        if seed is not None and not isinstance(seed, int):
            raise ValueError(f"expected an integer or None for seed but got {type(seed)}")
        if not isinstance(hash_function, str):
            raise ValueError(f"expected str for hash_function but got {type(hash_function)}")
        assert hash_function in [
            "murmurhash3",
            "xxhash",
            "sha1",
        ], f"hash_function must be one of 'murmurhash3', 'xxhash', 'sha1', got {hash_function}"

        return Series._from_pyseries(self._series.minhash(num_hashes, ngram_size, seed, hash_function))

    def _to_str_values(self) -> Series:
        return Series._from_pyseries(self._series.to_str_values())

    @property
    def float(self) -> SeriesFloatNamespace:
        return SeriesFloatNamespace.from_series(self)

    @property
    def str(self) -> SeriesStringNamespace:
        return SeriesStringNamespace.from_series(self)

    @property
    def dt(self) -> SeriesDateNamespace:
        return SeriesDateNamespace.from_series(self)

    @property
    def list(self) -> SeriesListNamespace:
        return SeriesListNamespace.from_series(self)

    @property
    def map(self) -> SeriesMapNamespace:
        return SeriesMapNamespace.from_series(self)

    @property
    def image(self) -> SeriesImageNamespace:
        return SeriesImageNamespace.from_series(self)

    @property
    def partitioning(self) -> SeriesPartitioningNamespace:
        return SeriesPartitioningNamespace.from_series(self)

    def __reduce__(
        self,
    ) -> (
        tuple[
            Callable[
                [
                    builtins.list[Any],
                    builtins.str,
                    DataType,
                ],
                Series,
            ],
            tuple[
                builtins.list[Any],
                builtins.str,
                DataType,
            ],
        ]
        | tuple[
            Callable[[pa.Array | pa.ChunkedArray, builtins.str, DataType], Series],
            tuple[pa.Array | pa.ChunkedArray, builtins.str, DataType],
        ]
    ):
        if self.datatype().is_python():
            return (Series.from_pylist, (self.to_pylist(), self.name(), self.datatype()))
        else:
            return (Series.from_arrow, (self.to_arrow(), self.name(), self.datatype()))

    def _debug_bincode_serialize(self) -> bytes:
        return self._series._debug_bincode_serialize()

    @classmethod
    def _debug_bincode_deserialize(cls, b: bytes) -> Series:
        return Series._from_pyseries(PySeries._debug_bincode_deserialize(b))

    def _eval_expressions(
        self,
        func_name: builtins.str,
        *others: Series | None,
        **kwargs: Any,
    ) -> Series:
        from daft.expressions.expressions import lit

        name = self._series.name()
        s = self._series
        other_series_list = []
        col_names = []
        for i, other in enumerate(others):
            col_name = f"c{i}"

            if other is None:
                continue
            if not isinstance(other, Series):
                try:
                    other_series_list.append(Series.from_pylist([other])._series.rename(col_name))
                    col_names.append(col_name)
                except AttributeError:
                    raise ValueError(f"expected another Series but got {type(other)}")
            else:
                other_series_list.append(other._series.rename(col_name))
                col_names.append(col_name)

        rb = PyRecordBatch.from_pyseries_list([s] + other_series_list)

        args = [native.unresolved_col(name)] + [native.unresolved_col(col_name) for col_name in col_names]

        f = native.get_function_from_registry(func_name)
        expr = f(
            *args,
            **{name: lit(v)._expr if not isinstance(v, PyExpr) else v for name, v in kwargs.items() if v is not None},
        ).alias(name)

        rb = rb.eval_expression_list([expr])
        pyseries = rb.get_column(0)
        return Series._from_pyseries(pyseries)


def item_to_series(name: str, item: Any) -> Series:
    # Fast path
    if isinstance(item, Series):
        return item

    if isinstance(item, list):
        series = Series.from_pylist(item, name)
    elif np.module_available() and isinstance(item, np.ndarray):  # type: ignore[attr-defined]
        series = Series.from_numpy(item, name)
    elif isinstance(item, (pa.Array, pa.ChunkedArray)):
        series = Series.from_arrow(item, name)
    elif pd.module_available() and isinstance(item, pd.Series):  # type: ignore[attr-defined]
        series = Series.from_pandas(item, name)
    else:
        raise ValueError(f"Creating a Series from data of type {type(item)} not implemented")
    return series


SomeSeriesNamespace = TypeVar("SomeSeriesNamespace", bound="SeriesNamespace")


class SeriesNamespace:
    _series: PySeries

    def __init__(self) -> None:
        raise NotImplementedError("We do not support creating a SeriesNamespace via __init__ ")

    @classmethod
    def from_series(cls: type[SomeSeriesNamespace], series: Series) -> SomeSeriesNamespace:
        ns = cls.__new__(cls)
        ns._series = series._series
        return ns

    def _eval_expressions(self, func_name: builtins.str, *others: Series | None, **kwargs: Any) -> Series:
        s = Series._from_pyseries(self._series)
        return s._eval_expressions(func_name, *others, **kwargs)


class SeriesFloatNamespace(SeriesNamespace):
    def is_nan(self) -> Series:
        return self._eval_expressions("is_nan")

    def is_inf(self) -> Series:
        return self._eval_expressions("is_inf")

    def not_nan(self) -> Series:
        return self._eval_expressions("not_nan")

    def fill_nan(self, fill_value: Series) -> Series:
        return self._eval_expressions("fill_nan", fill_value)


class SeriesStringNamespace(SeriesNamespace):
    def endswith(self, suffix: Series) -> Series:
        return self._eval_expressions("ends_with", suffix)

    def startswith(self, prefix: Series) -> Series:
        return self._eval_expressions("starts_with", prefix)

    def contains(self, pattern: Series) -> Series:
        return self._eval_expressions("utf8_contains", pattern)

    def match(self, pattern: Series) -> Series:
        return self._eval_expressions("regexp_match", pattern)

    def split(self, pattern: Series, regex: bool = False) -> Series:
        f_name = "regexp_split" if regex else "split"
        return self._eval_expressions(f_name, pattern)

    def concat(self, other: Series) -> Series:
        if not isinstance(other, Series):
            raise ValueError(f"expected another Series but got {type(other)}")
        assert self._series is not None and other._series is not None
        return Series._from_pyseries(self._series) + other

    def extract(self, pattern: Series, index: int = 0) -> Series:
        return self._eval_expressions("regexp_extract", pattern, index=index)

    def extract_all(self, pattern: Series, index: int = 0) -> Series:
        return self._eval_expressions("regexp_extract_all", pattern, index=index)

    def replace(self, pattern: Series, replacement: Series, regex: bool = False) -> Series:
        f_name = "regexp_replace" if regex else "replace"
        return self._eval_expressions(f_name, pattern, replacement)

    def length(self) -> Series:
        return self._eval_expressions("length")

    def length_bytes(self) -> Series:
        return self._eval_expressions("length_bytes")

    def lower(self) -> Series:
        return self._eval_expressions("lower")

    def upper(self) -> Series:
        return self._eval_expressions("upper")

    def lstrip(self) -> Series:
        return self._eval_expressions("lstrip")

    def rstrip(self) -> Series:
        return self._eval_expressions("rstrip")

    def reverse(self) -> Series:
        return self._eval_expressions("reverse")

    def capitalize(self) -> Series:
        return self._eval_expressions("capitalize")

    def left(self, nchars: Series) -> Series:
        return self._eval_expressions("left", nchars)

    def right(self, nchars: Series) -> Series:
        return self._eval_expressions("right", nchars)

    def find(self, substr: Series) -> Series:
        return self._eval_expressions("find", substr)

    def rpad(self, length: Series, pad: Series) -> Series:
        return self._eval_expressions("rpad", length, pad)

    def lpad(self, length: Series, pad: Series) -> Series:
        return self._eval_expressions("lpad", length, pad)

    def repeat(self, n: Series) -> Series:
        return self._eval_expressions("repeat", n)

    def like(self, pattern: Series) -> Series:
        return self._eval_expressions("like", pattern)

    def ilike(self, pattern: Series) -> Series:
        return self._eval_expressions("ilike", pattern)

    def to_date(self, format: str) -> Series:
        return self._eval_expressions("to_date", format=format)

    def to_datetime(self, format: str, timezone: str | None = None) -> Series:
        return self._eval_expressions("to_datetime", format=format, timezone=timezone)

    def substr(self, start: Series, length: Series | None = None) -> Series:
        return self._eval_expressions("substr", start, length)

    def normalize(
        self,
        *,
        remove_punct: bool = False,
        lowercase: bool = False,
        nfd_unicode: bool = False,
        white_space: bool = False,
    ) -> Series:
        return self._eval_expressions(
            "normalize",
            remove_punct=remove_punct,
            lowercase=lowercase,
            nfd_unicode=nfd_unicode,
            white_space=white_space,
        )

    def count_matches(
        self,
        patterns: Series,
        *,
        whole_words: bool = False,
        case_sensitive: bool = True,
    ) -> Series:
        pattern_expr = series_lit(patterns._series)
        return self._eval_expressions(
            "count_matches",
            patterns=pattern_expr,
            whole_words=whole_words,
            case_sensitive=case_sensitive,
        )


class SeriesDateNamespace(SeriesNamespace):
    def date(self) -> Series:
        return self._eval_expressions("date")

    def day(self) -> Series:
        return self._eval_expressions("day")

    def hour(self) -> Series:
        return self._eval_expressions("hour")

    def minute(self) -> Series:
        return self._eval_expressions("minute")

    def second(self) -> Series:
        return self._eval_expressions("second")

    def millisecond(self) -> Series:
        return self._eval_expressions("millisecond")

    def microsecond(self) -> Series:
        return self._eval_expressions("microsecond")

    def nanosecond(self) -> Series:
        return self._eval_expressions("nanosecond")

    def unix_date(self) -> Series:
        return self._eval_expressions("unix_date")

    def time(self) -> Series:
        return self._eval_expressions("time")

    def month(self) -> Series:
        return self._eval_expressions("month")

    def quarter(self) -> Series:
        return self._eval_expressions("quarter")

    def year(self) -> Series:
        return self._eval_expressions("year")

    def day_of_week(self) -> Series:
        return self._eval_expressions("day_of_week")

    def day_of_month(self) -> Series:
        return self._eval_expressions("day_of_month")

    def day_of_year(self) -> Series:
        return self._eval_expressions("day_of_year")

    def week_of_year(self) -> Series:
        return self._eval_expressions("week_of_year")

    def truncate(self, interval: str, relative_to: Series | None = None) -> Series:
        return self._eval_expressions("truncate", relative_to, interval=interval)

    def to_unix_epoch(self, time_unit: str | TimeUnit | None = None) -> Series:
        return self._eval_expressions("to_unix_epoch", time_unit=time_unit)

    def strftime(self, fmt: str | None = None) -> Series:
        return self._eval_expressions("strftime", format=fmt)

    def total_seconds(self) -> Series:
        return self._eval_expressions("total_seconds")

    def total_milliseconds(self) -> Series:
        return self._eval_expressions("total_milliseconds")

    def total_microseconds(self) -> Series:
        return self._eval_expressions("total_microseconds")

    def total_nanoseconds(self) -> Series:
        return self._eval_expressions("total_nanoseconds")

    def total_minutes(self) -> Series:
        return self._eval_expressions("total_minutes")

    def total_hours(self) -> Series:
        return self._eval_expressions("total_hours")

    def total_days(self) -> Series:
        return self._eval_expressions("total_days")


class SeriesPartitioningNamespace(SeriesNamespace):
    def days(self) -> Series:
        return Series._from_pyseries(self._series.partitioning_days())

    def hours(self) -> Series:
        return Series._from_pyseries(self._series.partitioning_hours())

    def months(self) -> Series:
        return Series._from_pyseries(self._series.partitioning_months())

    def years(self) -> Series:
        return Series._from_pyseries(self._series.partitioning_years())

    def iceberg_bucket(self, n: int) -> Series:
        return Series._from_pyseries(self._series.partitioning_iceberg_bucket(n))

    def iceberg_truncate(self, w: int) -> Series:
        return Series._from_pyseries(self._series.partitioning_iceberg_truncate(w))


class SeriesListNamespace(SeriesNamespace):
    def lengths(self) -> Series:
        warnings.warn(
            "This function will be deprecated from Daft version >= 0.3.5!  Instead, please use 'length'",
            category=DeprecationWarning,
        )

        return self._eval_expressions("list_count", mode=CountMode.All)

    def length(self) -> Series:
        return self._eval_expressions("list_count", mode=CountMode.All)

    def get(self, idx: Series, default: Series) -> Series:
        return self._eval_expressions("list_get", idx=idx, default=default)

    def sort(self, desc: bool | Series = False, nulls_first: bool | Series | None = None) -> Series:
        desc_expr = series_lit(desc._series) if isinstance(desc, Series) else desc
        return self._eval_expressions("list_sort", desc=desc_expr, nulls_first=nulls_first)


class SeriesMapNamespace(SeriesNamespace):
    def get(self, key: Series) -> Series:
        return Series._from_pyseries(self._series.map_get(key._series))


class SeriesImageNamespace(SeriesNamespace):
    def decode(
        self,
        on_error: Literal["raise", "null"] = "raise",
        mode: str | ImageMode | None = None,
    ) -> Series:
        return self._eval_expressions("image_decode", on_error=on_error, mode=mode)

    def encode(self, image_format: str | ImageFormat) -> Series:
        return self._eval_expressions("image_encode", image_format=image_format)

    def resize(self, w: int, h: int) -> Series:
        return self._eval_expressions("image_resize", w=w, h=h)

    def to_mode(self, mode: str | ImageMode) -> Series:
        if isinstance(mode, str):
            mode = ImageMode.from_mode_string(mode.upper())
        if not isinstance(mode, ImageMode):
            raise ValueError(f"mode must be a string or ImageMode variant, but got: {mode}")
        return self._eval_expressions("to_mode", mode=mode)



from __future__ import annotations

from typing import TYPE_CHECKING

from pyiceberg.io.pyarrow import schema_to_pyarrow
from pyiceberg.schema import Schema, SchemaVisitor

from daft import DataType
from daft.daft import PyField

if TYPE_CHECKING:
    from pyiceberg.types import ListType, MapType, NestedField, PrimitiveType, StructType

FieldIdMapping = dict[int, PyField]


def _nested_field_to_daft_pyfield(field: NestedField) -> PyField:
    return PyField.create(field.name, DataType.from_arrow_type(schema_to_pyarrow(field.field_type))._dtype)


class SchemaFieldIdMappingVisitor(SchemaVisitor[FieldIdMapping]):  # type: ignore[misc]
    """Extracts a mapping of {field_id: PyField} from an Iceberg schema."""

    def schema(self, schema: Schema, struct_result: FieldIdMapping) -> FieldIdMapping:
        """Visit a Schema."""
        return struct_result

    def struct(self, struct: StructType, field_results: list[FieldIdMapping]) -> FieldIdMapping:
        """Visit a StructType."""
        result = {field.field_id: _nested_field_to_daft_pyfield(field) for field in struct.fields}
        for r in field_results:
            result.update(r)
        return result

    def field(self, field: NestedField, field_result: FieldIdMapping) -> FieldIdMapping:
        """Visit a NestedField."""
        field_result[field.field_id] = _nested_field_to_daft_pyfield(field)
        return field_result

    def list(self, list_type: ListType, element_result: FieldIdMapping) -> FieldIdMapping:
        """Visit a ListType."""
        element_result[list_type.element_id] = _nested_field_to_daft_pyfield(list_type.element_field)
        return element_result

    def map(self, map_type: MapType, key_result: FieldIdMapping, value_result: FieldIdMapping) -> FieldIdMapping:
        result = {**key_result, **value_result}
        result[map_type.key_id] = _nested_field_to_daft_pyfield(map_type.key_field)
        result[map_type.value_id] = _nested_field_to_daft_pyfield(map_type.value_field)
        return result

    def primitive(self, primitive: PrimitiveType) -> FieldIdMapping:
        """Visit a PrimitiveType."""
        return {}



from __future__ import annotations

import pytest

PATHS = [
    "s3://eventual-dev-benchmarking-fixtures/parquet-benchmarking/tpch/1RG/daft_tpch_100g_32part_1RG.parquet",
    "s3://eventual-dev-benchmarking-fixtures/parquet-benchmarking/tpch/8RG/daft_tpch_100g_32part_8RG.parquet",
    "s3://eventual-dev-benchmarking-fixtures/parquet-benchmarking/tpch/64RG/daft_tpch_100g_32part_64RG.parquet",
]

IDS = ["1RG", "8RG", "64RG"]


@pytest.mark.benchmark(group="num_rowgroups_single_column")
@pytest.mark.parametrize(
    "path",
    PATHS,
    ids=IDS,
)
def test_read_parquet_num_rowgroups_single_column(path, read_fn, benchmark):
    data = benchmark(read_fn, path, columns=["L_ORDERKEY"])

    # Make sure the data is correct
    assert data.column_names == ["L_ORDERKEY"]
    assert len(data) == 18751674


@pytest.mark.benchmark(group="num_rowgroups_multi_contiguous_columns")
@pytest.mark.parametrize(
    "path",
    PATHS,
    ids=IDS,
)
def test_read_parquet_num_rowgroups_multi_contiguous_columns(path, read_fn, benchmark):
    data = benchmark(read_fn, path, columns=["L_ORDERKEY", "L_PARTKEY", "L_SUPPKEY"])

    # Make sure the data is correct
    assert data.column_names == ["L_ORDERKEY", "L_PARTKEY", "L_SUPPKEY"]
    assert len(data) == 18751674


@pytest.mark.benchmark(group="num_rowgroups_multi_sparse_columns")
@pytest.mark.parametrize(
    "path",
    PATHS,
    ids=IDS,
)
def test_read_parquet_num_rowgroups_multi_sparse_columns(path, read_fn, benchmark):
    data = benchmark(read_fn, path, columns=["L_ORDERKEY", "L_TAX"])

    # Make sure the data is correct
    assert data.column_names == ["L_ORDERKEY", "L_TAX"]
    assert len(data) == 18751674


@pytest.mark.benchmark(group="num_rowgroups_all_columns")
@pytest.mark.parametrize(
    "path",
    PATHS,
    ids=IDS,
)
def test_read_parquet_num_rowgroups_all_columns(path, read_fn, benchmark):
    data = benchmark(read_fn, path)

    # Make sure the data is correct
    assert len(data.column_names) == 16
    assert len(data) == 18751674



from __future__ import annotations

import re
from typing import Any

import numpy as np
import pyarrow as pa

from daft.recordbatch import RecordBatch

ANSI_ESCAPE = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")

TD_STYLE = 'style="text-align:left; max-width:192px; max-height:64px; overflow:auto"'
TH_STYLE = 'style="text-wrap: nowrap; max-width:192px; overflow:auto; text-align:left"'


def sort_arrow_table(tbl: pa.Table, *sort_by: str, ascending: bool = False) -> pa.Table:
    return tbl.sort_by([(name, "ascending" if ascending else "descending") for name in sort_by])


def assert_pyarrow_tables_equal(from_daft: pa.Table, expected: pa.Table) -> None:
    # Do a round-trip with Daft in order to cast pyarrow dtypes to Daft's supported Arrow dtypes (e.g. string -> large_string).
    expected = RecordBatch.from_arrow_table(expected).to_arrow_table()
    assert from_daft == expected, f"from_daft = {from_daft}\n\nexpected = {expected}"


def sort_pydict(pydict: dict[str, list[Any]], *sort_by: str, ascending: bool = False) -> pa.Table:
    return sort_arrow_table(RecordBatch.from_pydict(pydict).to_arrow_table(), *sort_by, ascending=ascending).to_pydict()


def random_numerical_embedding(
    rng: np.random.Generator, dtype: np.dtype, size: int, *, mult_for_int_like: int = 100
) -> np.ndarray:
    """Generate a uniformly distributed 1-D array of the specified datatype & dimensionality.

    The :param:`mult_for_int_like` parameter controls how much to multiply the randomly generated
    floating point values before rounding them to make integer vectors. Only applicable when
    :param:`dtype` is a signed or unsigned integer. Must be nonzero.

    :raises:`ValueError` if :param:`mult_for_int_like` is non-positive and :param:`dtype` is integer.
    """
    if dtype in (np.float32, np.float64):
        return rng.random(size=(size,), dtype=dtype)
    else:
        if mult_for_int_like <= 0:
            raise ValueError(
                "Multiplier for converting random float vectors into integer "
                f"vectors must be nonzero: {mult_for_int_like=}"
            )
        v = rng.random(size=(size,), dtype=np.float32)
        c = np.rint(v * mult_for_int_like)
        return c.astype(dtype)



from __future__ import annotations

import uuid

import pytest
import s3fs

import daft
from tests.conftest import get_tests_daft_runner_name, minio_create_public_bucket


@pytest.fixture(scope="function")
def bucket(minio_io_config):
    # For some reason s3fs is having trouble cleaning up MinIO
    # folders created by pyarrow write_parquet. We just write to
    # paths with random UUIDs to work around this.
    BUCKET = "my-bucket"

    fs = s3fs.S3FileSystem(
        key=minio_io_config.s3.key_id,
        password=minio_io_config.s3.access_key,
        client_kwargs={"endpoint_url": minio_io_config.s3.endpoint_url},
    )
    if not fs.exists(BUCKET):
        fs.mkdir(BUCKET)
    yield BUCKET


@pytest.mark.integration()
@pytest.mark.parametrize("protocol", ["s3://", "s3a://", "s3n://"])
def test_writing_parquet(minio_io_config, bucket, protocol):
    data = {
        "foo": [1, 2, 3],
        "bar": ["a", "b", "c"],
    }
    df = daft.from_pydict(data)
    df = df.repartition(2)
    results = df.write_parquet(
        f"{protocol}{bucket}/parquet-writes-{uuid.uuid4()}",
        partition_cols=["bar"],
        io_config=minio_io_config,
    )
    results.collect()
    assert len(results) == 3


@pytest.mark.integration()
@pytest.mark.parametrize("protocol", ["s3://", "s3a://", "s3n://"])
def test_writing_parquet_anonymous_mode(anonymous_minio_io_config, minio_io_config, protocol):
    bucket_name = "my-public-bucket"
    with minio_create_public_bucket(minio_io_config=minio_io_config, bucket_name=bucket_name):
        data = {
            "foo": [1, 2, 3],
            "bar": ["a", "b", "c"],
        }
        df = daft.from_pydict(data)
        df = df.repartition(2)
        results = df.write_parquet(
            f"{protocol}{bucket_name}/parquet-writes-{uuid.uuid4()}",
            partition_cols=["bar"],
            io_config=anonymous_minio_io_config,
        )
        results.collect()
        assert len(results) == 3


@pytest.mark.integration()
@pytest.mark.skipif(
    get_tests_daft_runner_name() != "native", reason="JSON writes are only implemented in the native runner"
)
@pytest.mark.parametrize("protocol", ["s3://", "s3a://", "s3n://"])
def test_writing_json(minio_io_config, bucket, protocol):
    data = {
        "foo": [1, 2, 3],
        "bar": ["a", "b", "c"],
    }
    df = daft.from_pydict(data)
    df = df.repartition(2)
    results = df.write_json(
        f"{protocol}{bucket}/json-writes-{uuid.uuid4()}",
        partition_cols=["bar"],
        io_config=minio_io_config,
    )
    results.collect()
    assert len(results) == 3



from __future__ import annotations

import pytest

import daft
from daft import col
from daft.sql import SQLCatalog


def test_simple_intersect(make_df):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"bar": [2, 3, 4]})
    expected = df1.intersect(df2).to_pydict()
    actual = daft.sql("select * from df1 intersect select * from df2").to_pydict()

    assert actual == expected


def test_intersect_with_duplicate(make_df):
    df1 = make_df({"foo": [1, 2, 2, 3]})
    df2 = make_df({"bar": [2, 3, 3]})
    expected = df1.intersect(df2).to_pydict()
    actual = daft.sql("select * from df1 intersect select * from df2").to_pydict()
    assert actual == expected


def test_self_intersect(make_df):
    df = make_df({"foo": [1, 2, 3]})
    expected = df.intersect(df).sort(by="foo")
    actual = daft.sql("select * from df intersect select * from df").sort(by="foo")
    assert actual.to_pydict() == expected.to_pydict()


def test_intersect_empty(make_df):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"bar": []}).select(col("bar").cast(daft.DataType.int64()))
    expected = df1.intersect(df2).to_pydict()
    actual = daft.sql("select * from df1 intersect select * from df2").to_pydict()
    assert actual == expected


def test_intersect_with_nulls(make_df):
    df1 = make_df({"foo": [1, 2, None]})
    df1_without_mull = make_df({"foo": [1, 2]})
    df2 = make_df({"bar": [2, 3, None]})
    df2_without_null = make_df({"bar": [2, 3]})

    expected = df1.intersect(df2)
    actual = daft.sql("select * from df1 intersect select * from df2")
    assert actual.to_pydict() == expected.to_pydict()

    expected = df1_without_mull.intersect(df2)
    actual = daft.sql("select * from df1_without_mull intersect select * from df2")
    assert actual.to_pydict() == expected.to_pydict()

    expected = df1.intersect(df2_without_null)
    actual = daft.sql("select * from df1 intersect select * from df2_without_null")
    assert actual.to_pydict() == expected.to_pydict()


def test_union_simple(make_df):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"foo": [2, 3, 4]})
    expected = df1.concat(df2).distinct().sort(by="foo").to_pydict()
    actual = daft.sql(
        "with unioned as (select * from df1 union select * from df2) select * from unioned order by foo"
    ).to_pydict()

    assert actual == expected


def test_union_all_simple(make_df):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"foo": [2, 3, 4]})
    expected = df1.concat(df2).to_pydict()
    actual = daft.sql("select * from df1 union all select * from df2").to_pydict()

    assert actual == expected


def test_union_with_nulls(make_df):
    df1 = make_df({"foo": [1, 2, 3]})
    df2 = make_df({"foo": [None]})

    expected = {"foo": [1, 2, 3, None]}
    catalog = SQLCatalog({"df1": df1, "df2": df2})
    actual = daft.sql("select * from df1 union select * from df2", catalog).to_pydict()
    assert actual == expected


def test_union_with_implicit_type_coercion(make_df):
    df1 = make_df({"foo": [1, 2, 3, 0, 1, 1]})
    df2 = make_df({"foo": ["hello", "1"]})
    catalog = SQLCatalog({"df1": df1, "df2": df2})
    expected = {"foo": ["0", "1", "2", "3", "hello"]}
    actual = daft.sql("select * from df1 union select * from df2", catalog).sort(by="foo").to_pydict()

    expected = {"foo": ["0", "1", "1", "1", "1", "2", "3", "hello"]}
    actual = daft.sql("select * from df1 union all select * from df2", catalog).sort(by="foo").to_pydict()
    assert actual == expected


def test_union_with_failed_implicit_type_coercion(make_df):
    df1 = make_df({"foo": [[1, 2, 3, 0, 1, 1], [None]]})
    df2 = make_df({"foo": ["hello", "1"]})
    catalog = SQLCatalog({"df1": df1, "df2": df2})

    with pytest.raises(Exception, match=r"unable to find a common supertype for union"):
        daft.sql("select * from df1 union all select * from df2", catalog).sort(by="foo").to_pydict()



from __future__ import annotations

import pyarrow as pa
import pytest

import daft
from daft import Catalog, Session
from daft.catalog import NotFoundError
from daft.logical.schema import DataType as dt
from daft.logical.schema import Field, Schema

CATALOG_ALIAS = "_test_catalog_iceberg"

# skip if pyarrow < 9
pyiceberg = pytest.importorskip("pyiceberg")
PYARROW_LOWER_BOUND_SKIP = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric()) < (9, 0, 0)
pytestmark = pytest.mark.skipif(PYARROW_LOWER_BOUND_SKIP, reason="iceberg not supported on old versions of pyarrow")


@pytest.fixture(scope="session")
def iceberg_catalog(tmp_path_factory):
    from pyiceberg.catalog.sql import SqlCatalog

    tmpdir = tmp_path_factory.mktemp("test_iceberg")
    catalog = SqlCatalog(
        CATALOG_ALIAS,
        **{
            "uri": f"sqlite:///{tmpdir}/iceberg_catalog.db",
            "warehouse": f"file://{tmpdir}",
        },
    )
    #
    # define a table via iceberg
    catalog.create_namespace("default")
    catalog.create_table(
        "default.tbl",
        pa.schema([("x", pa.bool_()), ("y", pa.int64()), ("z", pa.string())]),
    )
    return catalog


@pytest.fixture(scope="session")
def global_sess(iceberg_catalog):
    daft.attach_catalog(iceberg_catalog, alias=CATALOG_ALIAS)
    yield daft.current_session()
    daft.detach_catalog(alias=CATALOG_ALIAS)


@pytest.fixture(scope="session")
def sess(iceberg_catalog):
    sess = Session()
    sess.attach_catalog(iceberg_catalog, alias=CATALOG_ALIAS)
    yield sess
    sess.detach_catalog(alias=CATALOG_ALIAS)


@pytest.fixture(scope="session")
def catalog(iceberg_catalog):
    return Catalog.from_iceberg(iceberg_catalog)


def schema(fields: dict[str, dt]) -> Schema:
    return Schema._from_fields([Field.create(k, v) for k, v in fields.items()])


def assert_eq(df1, df2):
    assert df1.to_pydict() == df2.to_pydict()


def test_name(catalog):
    assert catalog.name == CATALOG_ALIAS
    assert catalog.get_table("default.tbl").name == "tbl"
    assert catalog.get_table("default.tbl").__repr__() == "Table('tbl')"


###
# ddl tests
###


def test_create_namespace(catalog: Catalog):
    c = catalog
    n = "test_create_namespace"
    #
    c.create_namespace(f"{n}")
    c.create_namespace(f"{n}.a")
    c.create_namespace(f"{n}.a.b")
    c.create_namespace(f"{n}.b")
    #
    # bug? iceberg sql catalog does not include child namespace
    # assert len(c.list_namespaces(f"{n}")) == 3
    assert len(c.list_namespaces(f"{n}.a")) == 1
    assert len(c.list_namespaces(f"{n}.b")) == 1

    # existence checks
    assert c.has_namespace(n)
    assert not c.has_namespace("x")

    #
    # err! should not exist
    with pytest.raises(Exception, match="does not exist"):
        c.list_namespaces("x")
    #
    # cleanup
    c.drop_namespace(n)


def test_create_namespace_if_exists(catalog: Catalog):
    c = catalog
    n = "test_create_namespace_if_exists"

    # test when namespace doesn't exist
    assert not c.has_namespace(n)
    c.create_namespace_if_not_exists(n)
    assert c.has_namespace(n)

    # should not raise an exception
    c.create_namespace_if_not_exists(n)
    assert c.has_namespace(n)

    # cleanup
    c.drop_namespace(n)


def test_create_table(catalog: Catalog):
    c = catalog
    n = "test_create_table"
    c.create_namespace(n)

    # create table with daft schema
    c.create_table(
        f"{n}.tbl1",
        schema(
            {
                "a": dt.bool(),
                "b": dt.int64(),
                "c": dt.string(),
            }
        ),
    )

    # test get_table
    assert c.get_table(f"{n}.tbl1")
    assert len(c.list_tables(n)) == 1

    # test has_table
    assert c.has_table(f"{n}.tbl1")
    assert not c.has_table(f"{n}.does_not_exist")

    # test exception
    with pytest.raises(NotFoundError):
        c.get_table(f"{n}.does_not_exist")

    # cleanup
    c.drop_table(f"{n}.tbl1")
    c.drop_namespace(n)


def test_create_table_if_not_exists(catalog: Catalog):
    c = catalog
    n = "test_create_table_if_not_exists"
    c.create_namespace(n)

    assert not c.has_table(f"{n}.tbl1")
    c.create_table_if_not_exists(
        f"{n}.tbl1",
        schema(
            {
                "a": dt.bool(),
                "b": dt.int64(),
                "c": dt.string(),
            }
        ),
    )
    assert c.has_table(f"{n}.tbl1")

    # should not raise an exception
    c.create_table_if_not_exists(
        f"{n}.tbl1",
        schema(
            {
                "a": dt.bool(),
                "b": dt.int64(),
                "c": dt.string(),
            }
        ),
    )
    assert c.has_table(f"{n}.tbl1")

    # Cleanup
    c.drop_table(f"{n}.tbl1")
    c.drop_namespace(n)


def test_create_table_as_select(catalog: Catalog):
    cat = catalog
    ns = "test_create_table_as_select"
    table_name = f"{ns}.tbl"
    df = daft.from_pydict({"a": [True, True, False], "b": [1, 2, 3], "c": ["x", "y", "z"]})

    # create_table with dataframe
    cat.create_namespace(ns)
    cat.create_table(table_name, df)

    # test get_table
    tbl = cat.get_table(table_name)
    assert tbl is not None
    assert len(cat.list_tables(ns)) == 1

    # validate the data
    assert tbl.read().to_pydict() == df.to_pydict()

    # cleanup
    cat.drop_table(table_name)
    cat.drop_namespace(ns)


def test_list_tables(catalog: Catalog, sess: Session):
    c = catalog

    # create a few namespaces
    n1 = "test_list_tables_a"
    n2 = "test_list_tables_b"
    n3 = "test_list_tables_c"
    c.create_namespace(n1)
    c.create_namespace(n2)
    c.create_namespace(n3)

    # no tables in any namespace
    assert len(c.list_tables()) == 1  # default
    assert len(c.list_tables(n1)) == 0
    assert len(c.list_tables(n2)) == 0
    assert len(c.list_tables(n3)) == 0

    # create a few tables in each namespace
    c.create_table(f"{n1}.tbl1_1", schema({"a": dt.bool()}))

    c.create_table(f"{n2}.tbl2_1", schema({"a": dt.bool()}))
    c.create_table(f"{n2}.tbl2_2", schema({"a": dt.bool()}))

    c.create_table(f"{n3}.tbl3_1", schema({"a": dt.bool()}))
    c.create_table(f"{n3}.tbl3_2", schema({"a": dt.bool()}))
    c.create_table(f"{n3}.tbl3_3", schema({"a": dt.bool()}))

    # list all tables
    assert len(c.list_tables()) == 7  # default + 6
    assert len(c.list_tables(n1)) == 1
    assert len(c.list_tables(n2)) == 2
    assert len(c.list_tables(n3)) == 3

    # show tables
    res = sess.sql("SHOW TABLES").to_pydict()
    assert len(res["table"]) == 7
    assert "tbl1_1" in res["table"]
    assert "tbl2_1" in res["table"]
    assert "tbl2_2" in res["table"]
    assert "tbl3_1" in res["table"]
    assert "tbl3_2" in res["table"]
    assert "tbl3_3" in res["table"]

    # # cleanup
    c.drop_table(f"{n1}.tbl1_1")
    c.drop_table(f"{n2}.tbl2_1")
    c.drop_table(f"{n2}.tbl2_2")
    c.drop_table(f"{n3}.tbl3_1")
    c.drop_table(f"{n3}.tbl3_2")
    c.drop_table(f"{n3}.tbl3_3")

    c.drop_namespace(n3)
    c.drop_namespace(n2)
    c.drop_namespace(n1)


###
# read tests
###


def test_daft_read_table(global_sess: Session):
    assert global_sess.read_table(f"{CATALOG_ALIAS}.default.tbl") is not None
    assert daft.read_table(f"{CATALOG_ALIAS}.default.tbl") is not None


def test_sess_read_table(sess: Session):
    # catalog-qualified
    assert sess.read_table(f"{CATALOG_ALIAS}.default.tbl") is not None
    # schema-qualified
    assert sess.read_table("default.tbl") is not None
    # unqualified
    sess.set_namespace("default")
    assert sess.read_table("tbl") is not None


def test_sess_read_options(sess: Session):
    # invalid options (version is a unity option, not iceberg)
    with pytest.raises(ValueError, match="Unsupported option"):
        sess.read_table("tbl", version="1")



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Generic, TypeVar

if TYPE_CHECKING:
    from collections.abc import Iterator

    from daft.recordbatch import MicroPartition
    from daft.schema import Schema

WriteResultType = TypeVar("WriteResultType")


@dataclass
class WriteResult(Generic[WriteResultType]):
    """Wrapper for result of the DataSink's `.write()` method.

    Attributes:
        result: The result from the write operation
        bytes_written: Size of the written data in bytes
        rows_written: Number of rows written
    """

    result: WriteResultType
    bytes_written: int
    rows_written: int


class DataSink(ABC, Generic[WriteResultType]):
    """Interface for writing data to a sink that is not built-in.

    When a DataFrame is written using the `.write_sink()` method, the following sequence occurs:

    1. The sink's `.start()` method is called once at the beginning of the write process.
    2. The DataFrame is executed, and its output is split into micropartitions.
    3. The sink's `.write()` method is invoked on each micropartition, potentially in parallel
       and distributed across multiple tasks or workers.
    4. After all writes complete, the resulting `WriteOutput` objects are gathered on a single node.
    5. The `.finalize()` method is then called with all write outputs to produce a final `MicroPartition`.

    Warning:
        This API is early in its development and is subject to change.
    """

    def name(self) -> str:
        """Optional custom sink name."""
        return "User-defined Data Sink"

    @abstractmethod
    def schema(self) -> Schema:
        """The expected schema for the micropartition returned by the `.finalize()` method of this DataSink.

        If this given schema does not match the actual schema of the micropartition at runtime, we throw an error.
        """
        raise NotImplementedError

    def start(self) -> None:
        """Optional callback for when a write operation begins.

        For example, this can be used to initialize resources, open connections, start a transaction etc.
        The default implementation does nothing.
        """
        pass

    @abstractmethod
    def write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[WriteResultType]]:
        """Writes a stream of micropartitions to the sink.

        This method should handle the ingestion of each micropartition and yield a result
        (e.g. metadata) for each successful write.

        Args:
            micropartitions (Iterator[MicroPartition]): An iterator of micropartitions to be written.

        Returns:
            Iterator[WriteResult[WriteResultType]]: An iterator of write results wrapped in a WriteOutput.
        """
        raise NotImplementedError

    def safe_write(self, micropartitions: Iterator[MicroPartition]) -> Iterator[WriteResult[WriteResultType]]:
        """This method wraps the abstract `write()` method with a try block to reraise potentially unserializable exceptions.

        Args:
            micropartitions (Iterator[MicroPartition]): An iterator of micropartitions to be written.

        Returns:
            Iterator[WriteResult[WriteResultType]]: An iterator of write results wrapped in a WriteOutput.

        Raises:
            Exception: Any exception that occurs during the write operation.
        """
        try:
            yield from self.write(micropartitions)
        except Exception as e:
            raise RuntimeError(f"Exception occurred while writing to {self.name()}: {type(e).__name__}: {e!s}") from e

    @abstractmethod
    def finalize(self, write_results: list[WriteResult[WriteResultType]]) -> MicroPartition:
        """Finalizes the write process and returns a resulting micropartition.

        For example, this can be used to merge, summarize, or commit the results of individual writes
        into a single output micropartition.

        Args:
            write_results (list[WriteResult[WriteResultType]]): The list of results from the calls to `.write()`.

        Returns:
            MicroPartition: A final, single micropartition representing the result of all writes.
        """
        raise NotImplementedError



DAFT_DASHBOARD_ENV_ENABLED: str
DAFT_DASHBOARD_ENV_NAME: str
DAFT_DASHBOARD_URL: str
DAFT_DASHBOARD_QUERIES_URL: str

class ConnectionHandle:
    def shutdown(self, noop_if_shutdown: bool) -> None: ...

def launch(detach: bool = False, noop_if_initialized: bool = False) -> ConnectionHandle: ...



from __future__ import annotations

import numpy as np
import pandas as pd
import pyarrow as pa
import pytest
from PIL import Image

import daft
from daft import DataType, Series, col
from daft.datatype import get_super_ext_type
from daft.utils import pyarrow_supports_fixed_shape_tensor

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())
DaftExtension = get_super_ext_type()


def test_embedding_type_df() -> None:
    data = [[1, 2, 3], np.arange(3), ["1", "2", "3"], [1, "2", 3.0], pd.Series([1.1, 2, 3]), (1, 2, 3), None]
    df = daft.from_pydict(
        {"index": np.arange(len(data)), "embeddings": Series.from_pylist(data, dtype=DataType.python())}
    )

    target = DataType.embedding(DataType.float32(), 3)
    df = df.select(col("index"), col("embeddings").cast(target))
    df = df.repartition(4, "index")
    df = df.sort("index")
    df = df.collect()
    arrow_table = df.to_arrow()
    assert isinstance(arrow_table["embeddings"].type, DaftExtension)


@pytest.mark.parametrize("from_pil_imgs", [True, False])
def test_image_type_df(from_pil_imgs) -> None:
    data = [
        np.arange(12, dtype=np.uint8).reshape((2, 2, 3)),
        np.arange(12, 39, dtype=np.uint8).reshape((3, 3, 3)),
        None,
    ]
    if from_pil_imgs:
        data = [Image.fromarray(arr, mode="RGB") if arr is not None else None for arr in data]
    df = daft.from_pydict({"index": np.arange(len(data)), "image": Series.from_pylist(data, pyobj="allow")})

    image_expr = col("image")
    if not from_pil_imgs:
        target = DataType.image("RGB")
        image_expr = image_expr.cast(target)
    df = df.select(col("index"), image_expr)
    df = df.repartition(4, "index")
    df = df.sort("index")
    df = df.collect()
    arrow_table = df.to_arrow()
    assert isinstance(arrow_table["image"].type, DaftExtension)


def test_fixed_shape_image_type_df() -> None:
    height = 2
    width = 2
    shape = (height, width, 3)
    data = [np.arange(12, dtype=np.uint8).reshape(shape), np.arange(12, 24, dtype=np.uint8).reshape(shape), None]
    df = daft.from_pydict({"index": np.arange(len(data)), "image": Series.from_pylist(data, dtype=DataType.python())})

    target = DataType.image("RGB", height, width)
    df = df.select(col("index"), col("image").cast(target))
    df = df.repartition(4, "index")
    df = df.sort("index")
    df = df.collect()
    arrow_table = df.to_arrow()
    assert isinstance(arrow_table["image"].type, DaftExtension)


def test_tensor_type_df() -> None:
    data = [
        np.arange(12).reshape((3, 2, 2)),
        np.arange(12, 39).reshape((3, 3, 3)),
        None,
    ]
    df = daft.from_pydict({"index": np.arange(len(data)), "tensor": Series.from_pylist(data, pyobj="allow")})

    df = df.select(col("index"), col("tensor"))
    df = df.repartition(4, "index")
    df = df.sort("index")
    df = df.collect()
    arrow_table = df.to_arrow()
    assert isinstance(arrow_table["tensor"].type, DaftExtension)


def test_fixed_shape_tensor_type_df() -> None:
    shape = (3, 2, 2)
    data = [
        np.arange(12).reshape(shape),
        np.arange(12, 24).reshape(shape),
        None,
    ]
    df = daft.from_pydict({"index": np.arange(len(data)), "tensor": Series.from_pylist(data, pyobj="allow")})

    target = DataType.tensor(DataType.int64(), shape)
    df = df.select(col("index"), col("tensor").cast(target))
    df = df.repartition(4, "index")
    df = df.sort("index")
    df = df.collect()
    arrow_table = df.to_arrow()
    if pyarrow_supports_fixed_shape_tensor():
        assert arrow_table["tensor"].type == pa.fixed_shape_tensor(pa.int64(), shape)
    else:
        assert isinstance(arrow_table["tensor"].type, DaftExtension)



from __future__ import annotations

import pytest

from daft import Expression, col, lit
from daft.dependencies import pa, pc


def assert_eq(actual: Expression, expect):
    """I don't know of a better way, but this can be switched out at a later time."""
    assert str(actual.to_arrow_expr()) == str(expect)


def test_col():
    assert_eq(col("a"), pc.field("a"))


def test_and():
    expr = lit(True) & lit(False)

    assert_eq(expr, "(true and false)")


def test_or():
    expr = lit(True) | lit(False)

    assert_eq(expr, "(true or false)")


def test_not():
    expr = ~lit(True)

    assert_eq(expr, "invert(true)")


def test_equal():
    expr = col("a") == col("b")

    assert_eq(expr, "(a == b)")


def test_not_equal():
    expr = col("a") != col("b")

    assert_eq(expr, "(a != b)")


def test_less_than():
    expr = col("a") < col("b")

    assert_eq(expr, "(a < b)")


def test_less_than_or_equal():
    expr = col("a") <= col("b")

    assert_eq(expr, "(a <= b)")


def test_greater_than():
    expr = col("a") > col("b")

    assert_eq(expr, "(a > b)")


def test_greater_than_or_equal():
    expr = col("a") >= col("b")

    assert_eq(expr, "(a >= b)")


def test_between():
    expr = col("a").between(col("b"), col("c"))

    assert_eq(expr, "((b <= a) and (a <= c))")


def test_is_in():
    expr = col("a").is_in([1, 2, 3])

    assert_eq(expr, pc.field("a").isin([1, 2, 3]))


def test_is_null():
    expr = col("a").is_null()

    assert_eq(expr, pc.field("a").is_null())


def test_not_null():
    expr = col("a").not_null()

    assert_eq(expr, ~pc.field("a").is_null())


def test_alias():
    expr = col("a").alias("b")

    assert_eq(expr, "a")


def test_cast():
    expr = col("a").cast("int64")

    assert_eq(expr, pc.field("a").cast(pa.int64()))


def test_lit():
    expr = lit(42)

    assert_eq(expr, "42")


@pytest.mark.parametrize(
    ["daft_expr", "arrow_expr"],
    [
        (lit(1).bitwise_and(lit(2)), "(1 and 2)"),
        (lit(1).bitwise_or(lit(2)), "(1 or 2)"),
        (lit(1) << lit(2), "shift_left(1, 2)"),
        (lit(1) >> lit(2), "shift_right(1, 2)"),
    ],
)
def test_bitwise_functions(daft_expr: Expression, arrow_expr: str):
    assert_eq(daft_expr, arrow_expr)


@pytest.mark.parametrize(
    ["daft_expr", "arrow_expr"],
    [
        (lit(1) + lit(2), "add(1, 2)"),
        (lit(1) - lit(2), "subtract(1, 2)"),
        (lit(1) * lit(2), "multiply(1, 2)"),
        (lit(1).ceil(), "ceil(1)"),
        (lit(1).floor(), "floor(1)"),
        (lit(1).sign(), "sign(1)"),
        (lit(1).signum(), "sign(1)"),
        (lit(1).negative(), "negate(1)"),
        (lit(1).round(2), pc.round(pc.scalar(1), 2)),
        (lit(1).sqrt(), "sqrt(1)"),
        (lit(1).sin(), "sin(1)"),
        (lit(1).cos(), "cos(1)"),
        (lit(1).tan(), "tan(1)"),
        (lit(1).arcsin(), "asin(1)"),
        (lit(1).arccos(), "acos(1)"),
        (lit(1).arctan(), "atan(1)"),
        (lit(1).arctan2(lit(2)), "atan2(1, 2)"),
        (lit(1).log2(), "log2(1)"),
        (lit(1).log10(), "log10(1)"),
        (lit(1).log(5), "logb(1, 5)"),
        (lit(1).ln(), "ln(1)"),
        (lit(1).log1p(), "log1p(1)"),
        # not available in older versions
        # (lit(1).exp(), "exp(1)"),
        # (lit(1).expm1(), "expm1(1)"),
        # not supported
        # (lit(1).cbrt(), "cbrt(1)"),
        # (lit(1).csc(), "csc(1)"),
        # (lit(1).sec(), "sec(1)"),
        # (lit(1).cot(), "cot(1)"),
        # (lit(1).sinh(), "sinh(1)"),
        # (lit(1).tanh(), "tanh(1)"),
        # (lit(1).arctanh(), "atanh(1)"),
        # (lit(1).arccosh(), "acosh(1)"),
        # (lit(1).arcsinh(), "asinh(1)"),
        # (lit(1).radians(), "radians(1)"),
        # (lit(1).degrees(), "degrees(1)"),
    ],
)
def test_math_functions(daft_expr: Expression, arrow_expr: str):
    assert_eq(daft_expr, arrow_expr)


@pytest.mark.parametrize(
    ["daft_expr", "arrow_expr"],
    [
        (lit("hello").str.capitalize(), 'utf8_capitalize("hello")'),
        (lit("hello").str.concat(lit("world")), 'add("hello", "world")'),
        # containment tests
        (lit("hello").str.count_matches(lit("l")), pc.count_substring(pc.scalar("hello"), "l")),
        (lit("hello").str.contains(lit("ll")), pc.match_substring(pc.scalar("hello"), "ll")),
        (lit("hello").str.endswith(lit("o")), pc.ends_with(pc.scalar("hello"), "o")),
        (lit("hello").str.startswith(lit("h")), pc.starts_with(pc.scalar("hello"), "h")),
        # TODO
        # (lit("hello").str.extract(lit("h(.*)o")), "extract('hello', 'h(.*)o')"),
        # (lit("hello").str.extract_all(lit("l")), "extract_all('hello', 'l')"),
        # (lit("hello").str.find(lit("l")), "find('hello', 'l')"),
        # (lit("hello").str.ilike(lit("HELLO")), "ilike('hello', 'HELLO')"),
        # (lit("hello").str.left(lit(3)), "left('hello', 3)"),
        # (lit("hello").str.length(), "utf8_length('hello')"),
        # (lit("hello").str.length_bytes(), "length_bytes('hello')"),
        # (lit("hello").str.like(lit("h%")), "like('hello', 'h%')"),
        # (lit("hello").str.lower(), "lower('hello')"),
        # (lit("hello").str.lpad(lit(10), lit(" ")), "lpad('hello', 10, ' ')"),
        # (lit("hello").str.lstrip(), "lstrip('hello')"),
        # (lit("hello").str.match(lit("h.*o")), "match('hello', 'h.*o')"),
        # (lit("hello").str.repeat(lit(3)), "repeat('hello', 3)"),
        # (lit("hello").str.replace(lit("l"), lit("L")), "replace('hello', 'l', 'L')"),
        # (lit("hello").str.reverse(), "reverse('hello')"),
        # (lit("hello").str.right(lit(3)), "right('hello', 3)"),
        # (lit("hello").str.rpad(lit(10), lit(" ")), "rpad('hello', 10, ' ')"),
        # (lit("hello").str.rstrip(), "rstrip('hello')"),
        # (lit("hello").str.split(lit("l")), "split('hello', 'l')"),
        # (lit("hello").str.substr(lit(1), lit(3)), "substr('hello', 1, 3)"),
        # (lit("hello").str.to_date(format="xxx"), "strftime('hello', 'xxx')"),
        # (lit("hello").str.to_datetime(format="xxx"), "strftime('hello', 'xxx')"),
        # (lit("hello").str.upper(), "utf8_upper('hello')"),
        # not supported
        # (lit("hello").str.normalize(lit("NFC")), "normalize('hello', 'NFC')"),
    ],
)
def test_string_functions(daft_expr: Expression, arrow_expr: str):
    assert_eq(daft_expr, arrow_expr)



from __future__ import annotations

import logging
import os
import pickle
import secrets
import subprocess
import sys
import tempfile
from multiprocessing import resource_tracker, shared_memory
from multiprocessing.connection import Listener
from typing import TYPE_CHECKING

from daft.errors import UDFException
from daft.expressions import Expression, ExpressionsProjection
from daft.recordbatch import MicroPartition

if TYPE_CHECKING:
    from daft.daft import PyExpr, PyMicroPartition

logger = logging.getLogger(__name__)

_ENTER = "__ENTER__"
_SUCCESS = "success"
_UDF_ERROR = "udf_error"
_ERROR = "error"
_SENTINEL = ("__EXIT__", 0)


class SharedMemoryTransport:
    def write_and_close(self, data: bytes) -> tuple[str, int]:
        shm = shared_memory.SharedMemory(create=True, size=len(data))
        # DO NOT REMOVE OR CHANGE THIS LINE. It is necessary to prevent the resource tracker from tracking the shared memory object.
        # This is because we are creating and unlinking the shared memory object in different processes.
        resource_tracker.unregister(shm._name, "shared_memory")  # type: ignore[attr-defined]
        shm.buf[: len(data)] = data
        shm.close()
        return shm.name, len(data)

    def read_and_release(self, name: str, size: int) -> bytes:
        shm = shared_memory.SharedMemory(name=name)
        try:
            data = bytes(shm.buf[:size])
        finally:
            shm.close()
            try:
                shm.unlink()
            except FileNotFoundError:
                logger.warning("Shared memory %s already unlinked", name)
        return data


class UdfHandle:
    def __init__(self, project_expr: PyExpr, passthrough_exprs: list[PyExpr]) -> None:
        # Construct UNIX socket path for basic communication
        with tempfile.NamedTemporaryFile(delete=True) as tmp:
            self.socket_path = tmp.name
        secret = secrets.token_bytes(32)
        self.listener = Listener(self.socket_path, authkey=secret)

        # Start the worker process
        self.process = subprocess.Popen(
            [
                sys.executable,
                "-m",
                "daft.execution.udf_worker",
                self.socket_path,
                secret.hex(),
            ]
        )

        # Initialize communication
        self.handle_conn = self.listener.accept()
        self.transport = SharedMemoryTransport()

        # Serialize and send the expression projection
        expr_projection = ExpressionsProjection(
            [Expression._from_pyexpr(expr) for expr in passthrough_exprs] + [Expression._from_pyexpr(project_expr)]
        )
        expr_projection_bytes = pickle.dumps(expr_projection)
        self.handle_conn.send((_ENTER, expr_projection_bytes))

    def eval_input(self, input: PyMicroPartition) -> PyMicroPartition:
        if self.process.poll() is not None:
            raise RuntimeError("UDF process has terminated")

        serialized = input.write_to_ipc_stream()
        shm_name, shm_size = self.transport.write_and_close(serialized)
        self.handle_conn.send((shm_name, shm_size))

        response = self.handle_conn.recv()
        if response[0] == _UDF_ERROR:
            base_exc: Exception = pickle.loads(response[3])
            if sys.version_info >= (3, 11):
                base_exc.add_note("\n".join(response[2].format()))
            raise UDFException(response[1]) from base_exc
        elif response[0] == _ERROR:
            raise RuntimeError("Actor Pool UDF unexpectedly failed with traceback:\n" + "\n".join(response[1].format()))
        elif response[0] == _SUCCESS:
            out_name, out_size = response[1], response[2]
            output_bytes = self.transport.read_and_release(out_name, out_size)
            deserialized = MicroPartition.from_ipc_stream(output_bytes)
            return deserialized._micropartition
        else:
            raise RuntimeError(f"Unknown response from actor: {response}")

    def teardown(self, timeout: float = 5.0) -> None:
        try:
            self.handle_conn.send(_SENTINEL)
        except (BrokenPipeError, EOFError):
            # If the connection is broken, just exit and join the process.
            pass
        self.handle_conn.close()
        self.listener.close()

        self.process.wait(timeout)
        if self.process.poll() is None:
            logger.warning("UDF did not shut down in time; terminating...")
            self.process.terminate()
            self.process.wait()

        if os.path.exists(self.socket_path):
            os.unlink(self.socket_path)



from __future__ import annotations

import pytest

import daft
from daft import col


def test_explode_basic():
    df = daft.from_pydict({"x": [[1, 2, 3], [4, 5], [], [6], None]})
    exploded = df.select(col("x").explode())
    result = exploded.to_pydict()["x"]
    # Should flatten all lists into rows, skipping empty lists
    assert result == [1, 2, 3, 4, 5, None, 6, None]


def test_explode_with_other_columns():
    df = daft.from_pydict({"id": [1, 2], "vals": [[10, 20], [30]]})
    exploded = df.select(col("id"), col("vals").explode())
    result = exploded.to_pydict()
    assert result["id"] == [1, 1, 2]
    assert result["vals"] == [10, 20, 30]


def test_explode_non_list_column():
    df = daft.from_pydict({"x": [1, 2, 3]})
    with pytest.raises(Exception):
        df.select(col("x").explode()).to_pydict()


def test_explode_multi_column():
    df = daft.from_pydict({"x": [[1, 2, 3], [4, 5], [], None], "y": [[6, 7, 8], [9, 10], [], None]})
    exploded = df.select(col("x").explode(), col("y").explode())
    result = exploded.to_pydict()
    # Should flatten, including None, and skip None rows
    assert result == {"x": [1, 2, 3, 4, 5, None, None], "y": [6, 7, 8, 9, 10, None, None]}


def test_explode_mismatched_column():
    df = daft.from_pydict({"x": [[1, 2, 3]], "y": [[4, 5]]})
    exploded = df.select(col("x").explode(), col("y").explode())

    with pytest.raises(Exception, match="DaftError::ValueError In multicolumn explode, list length did not match"):
        exploded.collect()



from __future__ import annotations

import pathlib

import pytest

import daft
from tests.integration.io.conftest import YieldFixture


@pytest.fixture(scope="function")
def local_image_data_fixture(tmpdir, image_data) -> YieldFixture[list[str]]:
    """Populates the local tmpdir with some fake data and returns filepaths."""
    # Dump some images into the tmpdir
    tmpdir = pathlib.Path(tmpdir)
    urls = []
    for i in range(10):
        path = tmpdir / f"{i}.jpeg"
        path.write_bytes(image_data)
        urls.append(str(path))

    yield urls

    # Cleanup tmpdir
    for child in tmpdir.glob("*"):
        child.unlink()


@pytest.mark.integration()
def test_url_download_local(local_image_data_fixture, image_data):
    data = {"urls": local_image_data_fixture}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download())
    assert df.to_pydict() == {**data, "data": [image_data for _ in range(len(local_image_data_fixture))]}


@pytest.mark.integration()
def test_url_download_local_missing(local_image_data_fixture):
    data = {"urls": local_image_data_fixture + ["/missing/path/x.jpeg"]}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(on_error="raise"))

    with pytest.raises(FileNotFoundError):
        df.collect()


@pytest.mark.integration()
def test_url_download_local_no_read_permissions(local_image_data_fixture, tmpdir):
    bad_permission_filepath = pathlib.Path(tmpdir) / "bad_file.jpeg"
    bad_permission_filepath.write_bytes(b"foo")
    bad_permission_filepath.chmod(0)

    data = {"urls": local_image_data_fixture + [str(bad_permission_filepath)]}
    df = daft.from_pydict(data)
    df = df.with_column("data", df["urls"].url.download(on_error="raise"))

    with pytest.raises(ValueError, match="Permission denied"):
        df.collect()



from __future__ import annotations

import numpy as np
import pyarrow as pa
import pytest

from daft.datatype import DataType
from daft.series import Series
from tests.series import ARROW_FLOAT_TYPES, ARROW_INT_TYPES
from tests.utils import ANSI_ESCAPE

ARROW_VERSION = tuple(int(s) for s in pa.__version__.split(".") if s.isnumeric())


@pytest.mark.parametrize("dtype", ARROW_INT_TYPES + ARROW_FLOAT_TYPES)
@pytest.mark.parametrize("use_offset_indices", [None, False, True])
def test_sparse_tensor_roundtrip(dtype, use_offset_indices):
    np_dtype = dtype.to_pandas_dtype()
    data = [
        np.array([[0, 1, 0, 0], [0, 0, 0, 0]], dtype=np_dtype),
        None,
        np.array([[0, 0, 0, 0], [0, 0, 0, 0]], dtype=np_dtype),
        np.array([[0, 0, 0, 0], [0, 0, 4, 0]], dtype=np_dtype),
    ]
    s = Series.from_pylist(data, pyobj="allow")

    tensor_dtype = DataType.tensor(DataType.from_arrow_type(dtype))

    t = s.cast(tensor_dtype)
    assert t.datatype() == tensor_dtype

    # Test sparse tensor roundtrip.
    if use_offset_indices is None:
        sparse_tensor_dtype = DataType.sparse_tensor(dtype=DataType.from_arrow_type(dtype))
    else:
        sparse_tensor_dtype = DataType.sparse_tensor(
            dtype=DataType.from_arrow_type(dtype), use_offset_indices=use_offset_indices
        )
    sparse_tensor_series = t.cast(sparse_tensor_dtype)
    assert sparse_tensor_series.datatype() == sparse_tensor_dtype
    back = sparse_tensor_series.cast(tensor_dtype)
    out = back.to_pylist()
    np.testing.assert_equal(out, data)


def test_sparse_tensor_repr():
    arr = np.arange(np.prod((2, 2)), dtype=np.int64).reshape((2, 2))
    arrs = [arr, arr, None]
    s = Series.from_pylist(arrs, pyobj="allow")
    s = s.cast(DataType.sparse_tensor(dtype=DataType.from_arrow_type(pa.int64())))
    out_repr = ANSI_ESCAPE.sub("", repr(s))
    assert (
        out_repr.replace("\r", "")
        == """╭────────────────────────────────────────────╮
│ list_series                                │
│ ---                                        │
│ SparseTensor[Int64; indices_offset: false] │
╞════════════════════════════════════════════╡
│ <SparseTensor shape=(2, 2)>                │
├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
│ <SparseTensor shape=(2, 2)>                │
├╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌╌┤
│ None                                       │
╰────────────────────────────────────────────╯
"""
    )


@pytest.mark.parametrize("indices_dtype", [np.uint8, np.uint16])
def test_minimal_indices_dtype_for_fixed_shape_sparse(indices_dtype: np.dtype):
    largest_index_possible = np.iinfo(indices_dtype).max
    tensor_shape = (largest_index_possible + 1, 1)

    series = Series.from_pylist([np.zeros(shape=tensor_shape)]).cast(
        DataType.tensor(DataType.float32(), shape=tensor_shape)
    )
    sparse_series = series.cast(DataType.sparse_tensor(DataType.float32(), shape=tensor_shape))

    received_tensor = sparse_series.to_pylist().pop()
    assert received_tensor["values"].dtype == np.float32
    assert received_tensor["indices"].dtype == indices_dtype
    assert received_tensor["shape"] == list(tensor_shape)



from __future__ import annotations

from daft.expressions import col
from daft.recordbatch import MicroPartition


def test_utf8_upper():
    table = MicroPartition.from_pydict({"col": ["Foo", None, "BarBaz", "quux", "1"]})
    result = table.eval_expression_list([col("col").str.upper()])
    assert result.to_pydict() == {"col": ["FOO", None, "BARBAZ", "QUUX", "1"]}



from __future__ import annotations

import uuid
from collections.abc import Generator
from typing import Any, Literal, Protocol

import pandas as pd
import pyarrow as pa
import pytest

import daft
import daft.context
from daft.recordbatch import MicroPartition

# import all conftest
from tests.integration.io.conftest import *


def pytest_addoption(parser):
    parser.addoption(
        "--credentials",
        action="store_true",
        help="Whether or not the current environment has access to remote storage credentials",
    )


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "integration: mark test as an integration test that runs with external dependencies"
    )


def get_tests_daft_runner_name() -> Literal["ray"] | Literal["native"]:
    """Test utility that checks the environment variable for the runner that is being used for the test."""
    name = os.getenv("DAFT_RUNNER")
    assert name is not None, "Tests must be run with $DAFT_RUNNER env var"
    name = name.lower()

    assert name in {"ray", "native"}, f"Runner name not recognized: {name}"
    return name


class UuidType(pa.ExtensionType):
    NAME = "daft.uuid"

    def __init__(self):
        pa.ExtensionType.__init__(self, pa.binary(), self.NAME)

    def __arrow_ext_serialize__(self):
        return b""

    def __reduce__(self):
        return UuidType, ()

    @classmethod
    def __arrow_ext_deserialize__(cls, storage_type, serialized):
        return cls()


@pytest.fixture
def uuid_ext_type() -> Generator[UuidType, None, None]:
    ext_type = UuidType()
    pa.register_extension_type(ext_type)
    yield ext_type
    pa.unregister_extension_type(ext_type.NAME)


@pytest.fixture(
    scope="function",
    params=[
        # Convert the data into Arrow and then load as in-memory Arrow data
        "arrow",
        # Dump the data as Parquet and load it as Parquet (will trigger "Unloaded" MicroPartitions)
        "parquet",
    ],
)
def data_source(request):
    return request.param


@pytest.fixture(scope="function")
def join_strategy(request):
    # Modifies the join strategy parametrization to toggle a specialized presorting path for sort-merge joins, where
    # each side of the join is sorted such that their boundaries will align.
    if request.param != "sort_merge_aligned_boundaries":
        yield request.param
    else:
        with daft.execution_config_ctx(sort_merge_join_sort_with_aligned_boundaries=True):
            yield "sort_merge"


@pytest.fixture(scope="function")
def make_spark_df(spark_session):
    def _make_spark_df(data: dict[str, Any]):
        fields = [name for name in data]
        rows = list(zip(*[data[name] for name in fields]))
        return spark_session.createDataFrame(rows, fields)

    yield _make_spark_df


@pytest.fixture(scope="function")
def assert_spark_equals(spark_session):
    def _assert_spark_dfs_eq(df1, df2):
        if isinstance(df1, daft.DataFrame):
            df1 = df1.to_pandas()
        else:
            df1 = df1.toPandas()
        if isinstance(df2, daft.DataFrame):
            df2 = df2.to_pandas()
        else:
            df2 = df2.toPandas()

        assert df1.equals(df2)

    yield _assert_spark_dfs_eq


class MakeDF(Protocol):
    def __call__(
        self, data: pa.Table | dict | list, repartition: int = 1, repartition_columns: list[str] = []
    ) -> daft.DataFrame: ...


@pytest.fixture(scope="function")
def make_df(data_source, tmp_path) -> Generator[MakeDF, None, None]:
    """Makes a dataframe when provided with data."""

    def _make_df(
        data: pa.Table | dict | list,
        repartition: int = 1,
        repartition_columns: list[str] = [],
    ) -> daft.DataFrame:
        pa_table: pa.Table
        if isinstance(data, pa.Table):
            pa_table = data
        elif isinstance(data, dict):
            pa_table = pa.table(data)
        elif isinstance(data, list):
            data = {k: [d[k] for d in data] for k in data[0].keys()}
            pa_table = pa.table(data)
        else:
            raise NotImplementedError(f"make_df not implemented for input type: {type(data)}")

        variant = data_source
        if variant == "arrow":
            df = daft.from_arrow(pa_table)
            if repartition != 1:
                return df.repartition(repartition, *repartition_columns)
            return df
        elif variant == "parquet":
            import pyarrow.parquet as papq

            name = str(uuid.uuid4())
            daft_recordbatch = MicroPartition.from_arrow(pa_table)
            partitioned_tables = (
                daft_recordbatch.partition_by_random(repartition, 0)
                if len(repartition_columns) == 0
                else daft_recordbatch.partition_by_hash([daft.col(c) for c in repartition_columns], repartition)
            )
            for i, tbl in enumerate(partitioned_tables):
                tmp_file = tmp_path / (name + f"-{i}")
                papq.write_table(tbl.to_arrow(), str(tmp_file))
            return daft.read_parquet(str(tmp_path) + f"/{name}-*")
        else:
            raise NotImplementedError(f"make_df not implemented for: {variant}")

    with daft.execution_config_ctx(
        # Disables merging of ScanTasks of Parquet when reading small Parquet files
        scan_tasks_min_size_bytes=0,
        scan_tasks_max_size_bytes=0,
    ):
        yield _make_df


def assert_df_equals(
    daft_df: pd.DataFrame,
    pd_df: pd.DataFrame,
    sort_key: str | list[str] = "Unique Key",
    assert_ordering: bool = False,
    check_dtype: bool = True,
):
    """Asserts that a Daft Dataframe is equal to a Pandas Dataframe.

    By default, we do not assert that the ordering is equal and will sort dataframes according to `sort_key`.
    However, if asserting on ordering is intended behavior, set `assert_ordering=True` and this function will
    no longer run sorting before running the equality comparison.
    """
    daft_pd_df = daft_df.reset_index(drop=True).reindex(sorted(daft_df.columns), axis=1)
    pd_df = pd_df.reset_index(drop=True).reindex(sorted(pd_df.columns), axis=1)

    # If we are not asserting on the ordering being equal, we run a sort operation on both dataframes using the provided sort key
    if not assert_ordering:
        sort_key_list: list[str] = [sort_key] if isinstance(sort_key, str) else sort_key
        for key in sort_key_list:
            assert key in daft_pd_df.columns, (
                f"Daft Dataframe missing key: {key}\nNOTE: This doesn't necessarily mean your code is "
                "breaking, but our testing utilities require sorting on this key in order to compare your "
                "Dataframe against the expected Pandas Dataframe."
            )
            assert key in pd_df.columns, (
                f"Pandas Dataframe missing key: {key}\nNOTE: This doesn't necessarily mean your code is "
                "breaking, but our testing utilities require sorting on this key in order to compare your "
                "Dataframe against the expected Pandas Dataframe."
            )
        daft_pd_df = daft_pd_df.sort_values(by=sort_key_list).reset_index(drop=True)
        pd_df = pd_df.sort_values(by=sort_key_list).reset_index(drop=True)

    assert sorted(daft_pd_df.columns) == sorted(pd_df.columns), f"Found {daft_pd_df.columns} expected {pd_df.columns}"
    for col in pd_df.columns:
        df_series = daft_pd_df[col]
        pd_series = pd_df[col]

        try:
            pd.testing.assert_series_equal(df_series, pd_series, check_dtype=check_dtype)
        except AssertionError:
            print(f"Failed assertion for col: {col}")
            raise


def check_answer(df: daft.DataFrame, expected_answer: dict[str, Any], is_sorted: bool = False):
    daft_df = df.to_pandas()
    expected_df = daft.from_pydict(expected_answer).to_pandas()
    # when this is an empty result, no need to check data types.
    check_dtype = not expected_df.empty
    if is_sorted:
        assert_df_equals(daft_df, expected_df, assert_ordering=True, check_dtype=check_dtype)
    else:
        sort_keys = df.column_names
        assert_df_equals(daft_df, expected_df, sort_key=sort_keys, assert_ordering=False, check_dtype=check_dtype)


@pytest.fixture(
    scope="function",
    params=[1, None] if get_tests_daft_runner_name() == "native" else [None],
)
def with_morsel_size(request):
    morsel_size = request.param
    with daft.context.execution_config_ctx(default_morsel_size=morsel_size):
        yield morsel_size



from __future__ import annotations

import pytest

import daft
from daft import DataType, col


def test_row_wise_udf():
    @daft.func()
    def my_stringify_and_sum(a: int, b: int) -> str:
        return f"{a + b}"

    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    actual = df.select(my_stringify_and_sum(col("x"), col("y"))).to_pydict()

    expected = {"x": ["5", "7", "9"]}

    assert actual == expected


def test_row_wise_udf_alternative_signature():
    @daft.func
    def my_stringify_and_sum(a: int, b: int) -> str:
        return f"{a + b}"

    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    actual = df.select(my_stringify_and_sum(col("x"), col("y"))).to_pydict()

    expected = {"x": ["5", "7", "9"]}

    assert actual == expected


def test_row_wise_udf_with_literal():
    @daft.func()
    def my_stringify_and_sum(a: int, b: int) -> str:
        return f"{a + b}"

    df = daft.from_pydict({"x": [1, 2, 3]})
    actual = df.select(my_stringify_and_sum(col("x"), 4)).to_pydict()

    expected = {"x": ["5", "6", "7"]}

    assert actual == expected


def test_row_wise_udf_should_infer_dtype_from_function():
    @daft.func()
    def list_string_return_type(a: int, b: int) -> list[str]:
        return [f"{a + b}"]

    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    df = df.select(list_string_return_type(col("x"), col("y")))

    schema = df.schema()
    expected_schema = daft.Schema.from_pydict({"x": daft.DataType.list(daft.DataType.string())})

    assert schema == expected_schema


def test_func_requires_return_dtype_when_no_annotation():
    with pytest.raises(ValueError, match="return_dtype is required when function has no return annotation"):

        @daft.func()
        def my_func(a: int, b: int):
            return f"{a + b}"


def test_row_wise_udf_override_return_dtype():
    @daft.func(return_dtype=DataType.int16())
    def my_sum(a: int, b: int) -> int:
        return a + b

    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    actual = df.select(my_sum(col("x"), col("y")))

    assert actual.schema()["x"].dtype == DataType.int16()

    expected = {"x": [5, 7, 9]}

    assert actual.to_pydict() == expected


def test_row_wise_udf_literal_eval():
    @daft.func
    def my_stringify_and_sum(a: int, b: int) -> str:
        return f"{a + b}"

    assert my_stringify_and_sum(1, 2) == "3"


def test_row_wise_udf_kwargs():
    @daft.func
    def my_stringify_and_sum_repeat(a: int, b: int, repeat: int = 1) -> str:
        return f"{a + b}" * repeat

    assert my_stringify_and_sum_repeat(1, 2) == "3"
    assert my_stringify_and_sum_repeat(1, 2, 3) == "333"

    df = daft.from_pydict({"x": [1, 2, 3], "y": [4, 5, 6]})
    default_df = df.select(my_stringify_and_sum_repeat(col("x"), col("y")))
    assert default_df.to_pydict() == {"x": ["5", "7", "9"]}

    constant_repeat_df = df.select(my_stringify_and_sum_repeat(col("x"), col("y"), repeat=2))
    assert constant_repeat_df.to_pydict() == {"x": ["55", "77", "99"]}

    dynamic_repeat_df = df.select(my_stringify_and_sum_repeat(col("x"), col("y"), repeat=col("x")))
    assert dynamic_repeat_df.to_pydict() == {"x": ["5", "77", "999"]}



from __future__ import annotations

from typing import Annotated

from fastapi import Header


def parse_range_from_header(range_header: Annotated[str, Header()], length: int) -> tuple[int, int]:
    """Parse the range header into a tuple of (start, end)."""
    start, end = (i for i in range_header[len("bytes=") :].split("-"))

    if start == "":  # suffix range: "bytes=-128"
        range_len = int(end)
        start = length - range_len if length > range_len else 0
        end = length - 1
    elif end == "":  # offset range: "bytes=128-"
        end = length - 1
        start = int(start)
    else:
        start = int(start)
        end = int(end)

    return start, end



from __future__ import annotations

import pickle
import sys
from multiprocessing.connection import Client
from traceback import TracebackException

from daft.errors import UDFException
from daft.execution.udf import _ENTER, _ERROR, _SENTINEL, _SUCCESS, _UDF_ERROR, SharedMemoryTransport
from daft.expressions.expressions import ExpressionsProjection
from daft.recordbatch.micropartition import MicroPartition


def udf_event_loop(
    secret: bytes,
    socket_path: str,
) -> None:
    # Initialize the client-side communication
    conn = Client(socket_path, authkey=secret)

    # Wait for the expression projection
    name, expr_projection_bytes = conn.recv()
    if name != _ENTER:
        raise ValueError(f"Expected '{_ENTER}' but got {name}")
    uninitialized_projection: ExpressionsProjection = pickle.loads(expr_projection_bytes)

    transport = SharedMemoryTransport()
    try:
        initialized_projection = ExpressionsProjection([e._initialize_udfs() for e in uninitialized_projection])

        while True:
            name, size = conn.recv()
            if (name, size) == _SENTINEL:
                break

            input_bytes = transport.read_and_release(name, size)
            input = MicroPartition.from_ipc_stream(input_bytes)
            evaluated = input.eval_expression_list(initialized_projection)
            output_bytes = evaluated.to_ipc_stream()

            out_name, out_size = transport.write_and_close(output_bytes)
            conn.send((_SUCCESS, out_name, out_size))
    except UDFException as e:
        exc = e.__cause__
        assert exc is not None
        conn.send((_UDF_ERROR, e.message, TracebackException.from_exception(exc), pickle.dumps(exc)))
    except Exception as e:
        try:
            conn.send((_ERROR, TracebackException.from_exception(e)))
        except Exception:
            # If the connection is broken, it's because the parent process has died.
            # We can just exit here.
            pass
    finally:
        conn.close()


if __name__ == "__main__":
    if len(sys.argv) != 3:
        print("Usage: python -m daft.execution.udf_worker <socket_path> <secret>", file=sys.stderr)
        sys.exit(1)

    socket_path = sys.argv[1]
    secret = bytes.fromhex(sys.argv[2])

    udf_event_loop(secret, socket_path)
